"""Boundary conditions for receptor based boundary conditions
including unbinding and rebinding.

"""

from collections import defaultdict
import logging
import itertools as it
import time

import numpy as np

from geomm.grouping import group_pair
from geomm.superimpose import superimpose
from geomm.rmsd import calc_rmsd
from geomm.centering import center_around

import mdtraj as mdj

from wepy.walker import WalkerState
from wepy.util.util import box_vectors_to_lengths_angles
from wepy.util.mdtraj import json_to_mdtraj_topology

from wepy.boundary_conditions.boundary import BoundaryConditions

class ReceptorBC(BoundaryConditions):
    """Abstract base class for ligand-receptor based boundary conditions.

    Provides shared utilities for warping walkers to any number of
    optionally weighted initial structures through a shared
    `warp_walkers` method.

    Non-abstract implementations of this class need only implement the
    `_progress` method which should return a boolean signalling a
    warping event and the dictionary-style warping record of the
    progress for only a single walker. These records will be collated
    into a single progress record across all walkers.

    Additionally, the `_update_bc` method can be overriden to return
    'BC' group records. That method should accept the arguments shown
    in this ABC and return a list of dictionary-style 'BC' records.

    Warping of walkers with multiple initial states will be done
    according to a choice of initial states weighted on their weights,
    if given.

    """

    # records of boundary condition changes (sporadic)
    BC_FIELDS = ()
    BC_SHAPES = ()
    BC_DTYPES = ()

    BC_RECORD_FIELDS = ()

    # warping fields are directly inherited

    # progress towards the boundary conditions (continual)
    PROGRESS_FIELDS = ()
    PROGRESS_SHAPES = ()
    PROGRESS_DTYPES = ()

    PROGRESS_RECORD_FIELDS = ()

    DISCONTINUITY_TARGET_IDXS = Ellipsis
    """Specifies which 'target_idxs' values are considered discontinuous targets.

    Values are either integer indices, Ellipsis (indicating all
    possible values are discontinuous), or None indicating no possible
    value is discontinuous.

    """

    def __init__(self, initial_states=None,
                 initial_weights=None,
                 ligand_idxs=None,
                 receptor_idxs=None,
                 **kwargs):
        """Base constructor for ReceptorBC.

        This should be called immediately in the subclass `__init__`
        method.

        If the initial weights for each initial state are not given
        uniform weights are assigned to them.

        Arguments
        ---------
        initial_states : list of objects implementing the State interface
            The list of possible states that warped walkers will assume.

        initial_weights : list of float, optional
            List of normalized probabilities of the initial_states
            provided. If not given, uniform probabilities will be
            used.

        ligand_idxs : arraylike of int
            The indices of the atom positions in the state considered
            the ligand.

        receptor_idxs : arraylike of int
            The indices of the atom positions in the state considered
            the receptor.

        Raises
        ------
        AssertionError
            If any of the following kwargs are not given:
            initial_states, ligand_idxs, receptor_idxs.

        """

        # make sure necessary inputs are given
        assert initial_states is not None, "Must give a set of initial states"
        assert ligand_idxs is not None, "Must give ligand indices"
        assert receptor_idxs is not None, "Must give binding site indices"

        self._initial_states = initial_states
        self._ligand_idxs = ligand_idxs
        self._receptor_idxs = receptor_idxs

        # we want to choose initial states conditional on their
        # initial probability if specified. If not specified assume
        # assume uniform probabilities.
        if initial_weights is None:
            self._initial_weights = [1/len(initial_states) for _ in initial_states]
        else:
            self._initial_weights = initial_weights

    @property
    def initial_states(self):
        """The possible initial states warped walkers may assume."""
        return self._initial_states

    @property
    def initial_weights(self):
        """The probabilities of each initial state being chosen during a warping."""
        return self._initial_weights

    @property
    def ligand_idxs(self):
        """The indices of the atom positions in the state considered the ligand."""
        return self._ligand_idxs

    @property
    def receptor_idxs(self):
        """The indices of the atom positions in the state considered the receptor."""

        return self._receptor_idxs


    def _progress(self, walker):
        """The method that must be implemented in non-abstract subclasses.

        Should decide if a walker should be warped or not and what its
        progress is regardless.

        Parameters
        ----------
        walker : object implementing the Walker interface

        Returns
        -------
        to_warp : bool
           Whether the walker should be warped or not.

        progress_data : dict of str : value
           Dictionary of the progress record group fields
           for this walker alone.

        """

        raise NotImplementedError

    def _warp(self, walker):
        """Perform the warping of a walker.

        Chooses an initial state to replace the walker's state with
        according to it's given weight.

        Returns a walker of the same type and weight.

        Parameters
        ----------
        walker : object implementing the Walker interface

        Returns
        -------
        warped_walker : object implementing the Walker interface
            The walker with the state after the warping. Weight should
            be the same.

        warping_data : dict of str : value
           The dictionary-style 'WARPING' record for this
           event. Excluding the walker index which is done in the main
           `warp_walkers` method.

        """


        # choose a state randomly from the set of initial states
        target_idx = np.random.choice(range(len(self.initial_states)), 1,
                                  p=self.initial_weights/np.sum(self.initial_weights))[0]

        warped_state = self.initial_states[target_idx]

        # set the initial state into a new walker object with the same weight
        warped_walker = type(walker)(state=warped_state, weight=walker.weight)

        # the data for the warp
        warp_data = {'target_idx' : np.array([target_idx]),
                     'weight' : np.array([walker.weight])}

        return warped_walker, warp_data


    def _update_bc(self, new_walkers, warp_data, progress_data, cycle):
        """Perform an update to the boundary conditions.

        No updates to the bc are ever done in this null
        implementation.

        Parameters
        ----------
        new_walkers : list of walkers
            The walkers after warping.

        warp_data : list of dict

        progress_data : dict

        cycle : int

        Returns
        -------
        bc_data : list of dict
            The dictionary-style records for BC update events

        """

        # do nothing by default
        return []


    def warp_walkers(self, walkers, cycle):
        """Test the progress of all the walkers, warp if required, and update
        the boundary conditions.

        Arguments
        ---------

        walkers : list of objects implementing the Walker interface

        cycle : int
            The index of the cycle.

        Returns
        -------

        new_walkers : list of objects implementing the Walker interface
            The new set of walkers that may have been warped.

        warp_data : list of dict of str : value
            The dictionary-style records for WARPING update events


        bc_data : list of dict of str : value
            The dictionary-style records for BC update events

        progress_data : dict of str : arraylike
            The dictionary-style records for PROGRESS update events

        """

        new_walkers = []

        # sporadic, zero or many records per call
        warp_data = []
        bc_data = []

        # continual, one record per call
        progress_data = defaultdict(list)

        # calculate progress data
        all_progress_data = [self._progress(w) for w in walkers]

        for walker_idx, walker in enumerate(walkers):

            # unpack progress data
            to_warp, walker_progress_data = all_progress_data[walker_idx]

            # add that to the progress data record
            for key, value in walker_progress_data.items():
                progress_data[key].append(value)

            # if the walker is meets the requirements for warping warp
            # it
            if to_warp:
                # warp the walker
                warped_walker, walker_warp_data = self._warp(walker)

                # add the walker idx to the walker warp record
                walker_warp_data['walker_idx'] = np.array([walker_idx])

                # save warped_walker in the list of new walkers to return
                new_walkers.append(warped_walker)

                # save the instruction record of the walker
                warp_data.append(walker_warp_data)

                logging.info('WARP EVENT observed at {}'.format(cycle))
                logging.info('Warped Walker Weight = {}'.format(
                    walker_warp_data['weight']))

            # no warping so just return the original walker
            else:
                new_walkers.append(walker)

        # consolidate the progress data to an array of a single
        # feature vectors for the cycle
        for key, value in progress_data.items():
            progress_data[key] = value

        # if the boundary conditions need to be updated given the
        # cycle and state from warping perform that now and return any
        # record data for that
        bc_data = self._update_bc(new_walkers, warp_data, progress_data, cycle)

        return new_walkers, warp_data, bc_data, progress_data

    @classmethod
    def warping_discontinuity(cls, warping_record):
        """Tests whether a warping record generated by this class is
        discontinuous or not.

        Parameters
        ----------

        warping_record : tuple
            The WARPING type record.

        Returns
        -------

        is_discontinuous : bool
            True if a discontinuous warp False if continuous.

        """

        # if it is Ellipsis then all possible values are discontinuous
        if cls.DISCONTINUITY_TARGET_IDXS is Ellipsis:
            return True

        # if it is None then all possible values are continuous
        elif cls.DISCONTINUITY_TARGET_IDXS is None:
            return False

        # otherwise it will have a tuple of indices for the
        # target_idxs that are discontinuous targets
        elif warping_record[2] in cls.DISCONTINUITY_TARGET_IDXS:
            return True

        # otherwise it wasn't a discontinuous target
        else:
            return False


class RebindingBC(ReceptorBC):
    """Boundary condition for doing re-binding simulations of ligands to a
    receptor.

    Implements the ReceptorBC superclass.

    This boundary condition will warp walkers to a number of initial
    states whenever a walker becomes very close to the native (bound)
    state.

    Thus the choice of the 'initial_states' argument should be walkers
    which are completely unbound (the choice of which are weighted by
    'initial_weight') and the choice of 'native_state' should be of a
    ligand bound to the receptor, e.g. X-ray crystallography or docked
    structure.

    The cutoff for the boundary is an RMSD of the walker to the native
    state which is calculated by first aligning and superimposing the
    entire structure according the atom indices specified in
    'binding_site_idxs', and as the name suggests should correspond to
    some approximation of the binding site of the ligand that occurs
    in the native state. Then the raw RMSD of the native and walker
    ligands is calculated. If this RMSD is less than the 'cutoff_rmsd'
    argument the walker is warped.

    PROGRESS is reported for each walker from this rmsd.

    The BC records are never updated.

    """

    # Records of boundary condition changes (sporadic)
    BC_FIELDS = ReceptorBC.BC_FIELDS + ('native_rmsd_cutoff', )
    """The 'native_rmsd_cutoff' is the cutoff used to determine when
    walkers have re-bound to the receptor, which is defined as the
    RMSD of the ligand to the native ligand bound state, when the
    binding sites are aligned and superimposed.

    """

    BC_SHAPES = ReceptorBC.BC_SHAPES + ((1,), )
    BC_DTYPES = ReceptorBC.BC_DTYPES + (np.float, )

    BC_RECORD_FIELDS = ReceptorBC.BC_RECORD_FIELDS + ('native_rmsd_cutoff', )

    # warping (sporadic)
    WARPING_FIELDS = ReceptorBC.WARPING_FIELDS + ()
    WARPING_SHAPES = ReceptorBC.WARPING_SHAPES + ()
    WARPING_DTYPES = ReceptorBC.WARPING_DTYPES + ()

    WARPING_RECORD_FIELDS = ReceptorBC.WARPING_RECORD_FIELDS + ()

    # progress towards the boundary conditions (continual)
    PROGRESS_FIELDS = ReceptorBC.PROGRESS_FIELDS + ('native_rmsd',)
    PROGRESS_SHAPES = ReceptorBC.PROGRESS_SHAPES + (Ellipsis,)
    PROGRESS_DTYPES = ReceptorBC.PROGRESS_DTYPES + (np.float,)

    PROGRESS_RECORD_FIELDS = ReceptorBC.PROGRESS_RECORD_FIELDS + ('native_rmsd', )
    """Records for the state of this record group.

    The 'native_rmsd' is the is the RMSD of the ligand to the native
    ligand bound state, when the binding sites are aligned and
    superimposed.

    """

    def __init__(self, native_state=None,
                 cutoff_rmsd=0.2,
                 initial_states=None,
                 initial_weights=None,
                 ligand_idxs=None,
                 binding_site_idxs=None,
                 **kwargs):
        """Constructor for RebindingBC.

        Arguments
        ---------

        native_state : object implementing the State interface
            The reference bound state. Will be automatically centered.

        cutoff_rmsd : float
            The cutoff RMSD for considering a walker bound.

        initial_states : list of objects implementing the State interface
            The list of possible states that warped walkers will assume.

        initial_weights : list of float, optional
            List of normalized probabilities of the initial_states
            provided. If not given, uniform probabilities will be
            used.

        ligand_idxs : arraylike of int
            The indices of the atom positions in the state considered
            the ligand.

        binding_site_idxs : arraylike of int
            The indices of the atom positions in the state considered
            the binding site.

        Raises
        ------
        AssertionError
            If any of the following kwargs are not given:
            native_state, initial_states, ligand_idxs, receptor_idxs.


        """

        super().__init__(initial_states=initial_states,
                         initial_weights=initial_weights,
                         ligand_idxs=ligand_idxs,
                         receptor_idxs=binding_site_idxs
                         **kwargs)

        # test inputs
        assert native_state is not None, "Must give a native state"
        assert type(cutoff_rmsd) is float

        native_state_d = native_state.dict()

        # save the native state and center it around it's binding site
        native_state_d['positions'] = center_around(native_state['positions'], binding_site_idxs)

        native_state = WalkerState(**native_state_d)

        # save attributes
        self._native_state = native_state
        self._cutoff_rmsd = cutoff_rmsd

    @property
    def native_state(self):
        """The reference bound state to which walkers are compared."""
        return self._native_state

    @property
    def cutoff_rmsd(self):
        """The cutoff RMSD for considering a walker bound."""
        return self._cutoff_rmsd

    @property
    def binding_site_idxs(self):
        """The indices of the atom positions in the state considered the binding site."""

        return self._receptor_idxs

    def _progress(self, walker):
        """Calculate if the walker has bound and provide progress record.

        Parameters
        ----------
        walker : object implementing the Walker interface

        Returns
        -------
        is_bound : bool
           Whether the walker is unbound (warped) or not

        progress_data : dict of str : value
           Dictionary of the progress record group fields
           for this walker alone.

        """

        # first recenter the ligand and the receptor in the walker
        box_lengths, box_angles = box_vectors_to_lengths_angles(walker.state['box_vectors'])
        grouped_walker_pos = group_pair(walker.state['positions'], box_lengths,
                                     self.binding_site_idxs, self.ligand_idxs)

        # center the positions around the center of the binding site
        centered_walker_pos = center_around(grouped_walker_pos, self.binding_site_idxs)

        # superimpose the walker state positions over the native state
        # matching the binding site indices only
        sup_walker_pos, _, _ = superimpose(self.native_state['positions'], centered_walker_pos,
                                 idxs=self.binding_site_idxs)

        # calculate the rmsd of the walker ligand (superimposed
        # according to the binding sites) to the native state ligand
        native_rmsd = calc_rmsd(self.native_state['positions'], sup_walker_pos,
                                idxs=self.ligand_idxs)

        # test to see if the ligand is re-bound
        rebound = False
        if native_rmsd <= self.cutoff_rmsd:
            rebound = True

        progress_data = {'native_rmsd' : native_rmsd}

        return rebound, progress_data


class UnbindingBC(ReceptorBC):
    """Boundary condition for ligand unbinding.

    Walkers will be warped (discontinuously) if all atoms in the
    ligand are at least a certain distance away from the atoms in the
    receptor (i.e. the min-min of ligand-receptor distances > cutoff).

    Warping will replace the walker state with the initial state given
    as a parameter to this class.

    Also reports on the progress of that min-min for each walker.

    """

    # records of boundary condition changes (sporadic)
    BC_FIELDS = ReceptorBC.BC_FIELDS + ('boundary_distance', )
    """
    Only occurs at the start of the simulation and just reports on the
    min-min cutoff distance.
    """

    BC_SHAPES = ReceptorBC.BC_SHAPES + ((1,), )
    BC_DTYPES = ReceptorBC.BC_DTYPES + (np.float, )
    BC_RECORD_FIELDS = ReceptorBC.BC_RECORD_FIELDS + ('boundary_distance', )

    # warping (sporadic)
    WARPING_FIELDS = ReceptorBC.WARPING_FIELDS + ()
    WARPING_SHAPES = ReceptorBC.WARPING_SHAPES + ()
    WARPING_DTYPES = ReceptorBC.WARPING_DTYPES + ()

    WARPING_RECORD_FIELDS = ReceptorBC.WARPING_RECORD_FIELDS + ()

    # progress record group
    PROGRESS_FIELDS = ReceptorBC.PROGRESS_FIELDS + ('min_distances',)
    """
    The 'min_distances' field reports on the min-min ligand-receptor
    distance for each walker.

    """

    PROGRESS_SHAPES = ReceptorBC.PROGRESS_SHAPES + (Ellipsis,)
    PROGRESS_DTYPES = ReceptorBC.PROGRESS_DTYPES + (np.float,)
    PROGRESS_RECORD_FIELDS = ReceptorBC.PROGRESS_RECORD_FIELDS + ('min_distances', )

    def __init__(self, initial_state=None,
                 cutoff_distance=1.0,
                 topology=None,
                 ligand_idxs=None,
                 receptor_idxs=None,
                 periodic=True,
                 **kwargs):
        """Constructor for UnbindingBC class.

        All the key-word arguments are necessary.

        The 'initial_state' should be the initial state of your
        simulation for proper non-equilibrium simulations.

        Arguments
        ---------
        initial_state : object implementing State interface
            The state walkers will take on after unbinding.

        cutoff_distance : float
            The distance that specifies the boundary condition. When
            the min-min ligand-receptor distance is less than this it
            will be warped.

        topology : str
            A JSON string of topology.

        ligand_idxs : list of int
           Indices of the atoms in the topology that correspond to the ligands.

        receptor_idxs : list of int
           Indices of the atoms in the topology that correspond to the
           receptor for the ligand.

        Raises
        ------
        AssertionError
            If any of the following are not provided: initial_state, topology,
            ligand_idxs, receptor_idxs

        AssertionError
            If the cutoff distance is not a float.

        Warnings
        --------
        The 'initial_state' should be the initial state of your
        simulation for proper non-equilibrium simulations.

        Notes
        -----

        The topology argument is necessary due to an implementation
        detail that uses mdtraj and may not be required in the future.

        """

        # since the super class can handle multiple initial states we
        # wrap the single initial state to a list.
        super().__init__(initial_states=[initial_state],
                         ligand_idxs=ligand_idxs,
                         receptor_idxs=receptor_idxs,
                         **kwargs)

        # test input
        assert topology is not None, "Must give a reference topology"
        assert type(cutoff_distance) is float

        self._cutoff_distance = cutoff_distance
        self._topology = topology

        # convert the json topology to an mdtraj one
        self._mdj_top = json_to_mdtraj_topology(self._topology)

        # whether or not to use the periodic box vectors in the
        # distance calculation
        self._periodic = periodic


    @property
    def cutoff_distance(self):
        """The distance a ligand must be to be unbound."""
        return self._cutoff_distance

    @property
    def topology(self):
        """JSON string topology of the system."""
        return self._topology

    def _calc_min_distance(self, walker):
        """Min-min distance for a walker.

        Parameters
        ----------
        walker : object implementing the Walker interface

        Returns
        -------
        min_distance : float

        """

        cell_lengths, cell_angles = box_vectors_to_lengths_angles(walker.state['box_vectors'])

        t2 = time.time()
        # make a traj out of it so we can calculate distances through
        # the periodic boundary conditions
        walker_traj = mdj.Trajectory(walker.state['positions'],
                                     topology=self._mdj_top,
                                     unitcell_lengths=cell_lengths,
                                     unitcell_angles=cell_angles)

        t3 = time.time()
        # calculate the distances through periodic boundary conditions
        # and get hte minimum distance
        min_distance = np.min(mdj.compute_distances(walker_traj,
                                                    it.product(self.ligand_idxs,
                                                               self.receptor_idxs),
                                                    periodic=self._periodic)
        )
        t4 = time.time()
        logging.info("Make a traj: {0}; Calc dists: {1}".format(t3-t2,t4-t3))

        return min_distance

    def _progress(self, walker):
        """Calculate whether a walker has unbound and also provide a
        dictionary for a single walker in the progress records.

        Parameters
        ----------
        walker : object implementing the Walker interface

        Returns
        -------
        is_unbound : bool
           Whether the walker is unbound (warped) or not

        progress_data : dict of str : value
           Dictionary of the progress record group fields
           for this walker alone.

        """

        min_distance = self._calc_min_distance(walker)

        # test to see if the ligand is unbound
        unbound = False
        if min_distance >= self._cutoff_distance:
            unbound = True

        progress_data = {'min_distances' : min_distance}

        return unbound, progress_data

    def _update_bc(self, new_walkers, warp_data, progress_data, cycle):
        """Perform an update to the boundary conditions.

        This is only used on the first cycle to keep a record of the
        cutoff parameter.

        Parameters
        ----------
        new_walkers : list of walkers
            The walkers after warping.

        warp_data : list of dict

        progress_data : dict

        cycle : int

        Returns
        -------
        bc_data : list of dict
            The dictionary-style records for BC update events

        """

        # Only report a record on
        # the first cycle which gives the distance at which walkers
        # are warped
        if cycle == 0:
            return [{'boundary_distance' : np.array([self._cutoff_distance]),},]
        else:
            return []



from wepy_tools.sim_makers.openmm.sim_maker import OpenMMSimMaker, OpenMMToolsTestSysSimMaker


from wepy_tools.sim_makers.openmm.lennard_jones import LennardJonesPairOpenMMSimMaker
from wepy_tools.sim_makers.openmm.lysozyme import LysozymeImplicitOpenMMSimMaker

__all__ = [
    "LennardJonesPairOpenMMSimMaker",
    "LysozymeImplicitOpenMMSimMaker",
] + [ # the base classes
    "OpenMMSimMaker", "OpenMMToolsTestSysSimMaker"
]




if __name__ == "__main__":

    import os
    import shutil
    import sys
    import logging
    from pathlib import Path

    from multiprocessing_logging import install_mp_handler

    from wepy_tools.sim_makers.openmm.lysozyme import LysozymeImplicitOpenMMSimMaker

    logging.getLogger().setLevel(logging.DEBUG)
    install_mp_handler()

    if sys.argv[1] == "-h" or sys.argv[1] == "--help":
        print("arguments: n_cycles, n_steps, n_walkers, n_workers, platform, resampler")
        exit()
    else:
        n_cycles = int(sys.argv[1])
        n_steps = int(sys.argv[2])
        n_walkers = int(sys.argv[3])
        n_workers = int(sys.argv[4])
        platform = sys.argv[5]
        resampler = sys.argv[6]

        print("Number of steps: {}".format(n_steps))
        print("Number of cycles: {}".format(n_cycles))


    output_dir = Path('_output')

    # make the results directory if not already made
    try:
        shutil.rmtree(output_dir / 'we')
    except FileNotFoundError:
        pass

    os.makedirs(output_dir / 'we', exist_ok=True)

    sim_maker = LysozymeImplicitOpenMMSimMaker()

    apparatus = sim_maker.make_apparatus(
        integrator='LangevinIntegrator',
        resampler=resampler,
        bc='UnbindingBC',
        platform=platform,
    )
    config = sim_maker.make_configuration(apparatus,
                                          work_mapper_spec='TaskMapper',
                                          platform=platform,
                                          work_dir=str(output_dir / 'we'))

    sim_manager = sim_maker.make_sim_manager(n_walkers, apparatus, config)

    sim_manager.run_simulation(n_cycles, n_steps, num_workers=n_workers)



import numpy as np
from scipy.spatial.distance import euclidean

from wepy.resampling.distances.distance import Distance

class PairDistance(Distance):

    def __init__(self, metric=euclidean):
        self.metric = metric

    def image(self, state):
        return state['positions']

    def image_distance(self, image_a, image_b):
        dist_a = self.metric(image_a[0], image_a[1])
        dist_b = self.metric(image_b[0], image_b[1])

        return np.abs(dist_a - dist_b)



"""Collection of routines for counting transitions and calculating
transition probability matrices.

There are two basic data structures used in this module for edge
transitions: dictionary based and matrix based.

The dictionary based approach is used to interface with graph
representations, like in the
~wepy.analysis.network.MacroStateNetwork~.

Arrays are used as an interface to numerical libraries which use this
to calculate other properties such as steady-state probabilities.

Wepy is primarily meant to support expedient data structuring and will
not support the latter numerical calculations and is left up to
external libraries and domain specific code.

Routines
--------

transition_counts : Given a structured array of assignments of
    microstate to macrostate labels and a listing of all kinetic
    transitions over the microstates returns the counts of transitions
    between each macrostate as a dictionary mapping macrostate
    transition edges to the total counts each microstate transition.

counts_d_to_matrix : Converts a dictionary mapping macrostate
    transition edges to a asymmetric transition matrix.

normalize_counts : Normalizes outgoing transition counts from each
    macrostate to 1 for a transition counts matrix.

transition_counts_matrix : Given a structured array of assignments of
    microstate to macrostate labels and a listing of all kinetic
    transitions over the microstates returns the counts of transitions
    between each macrostate as an assymetric transition counts matrix.

transition_probability_matrix : Given a structured array of assignments of
    microstate to macrostate labels and a listing of all kinetic
    transitions over the microstates returns the probability of transitions
    between each macrostate as an assymetric transition probability matrix.

run_transition_counts_matrix : Generates an asymmetric transition
    counts matrix directly from a single WepyHDF5 run.

run_transition_probability_matrix : Generates an asymmetric transition
    probability matrix directly from a single WepyHDF5 run.
"""

import itertools as it
from collections import defaultdict

import numpy as np

def transition_counts(
        assignments,
        transitions,
        weights=None):
    """Make a dictionary of the count of microstate transitions between macrostates.

    If weights are given counts are the weight instead of 1.

    Parameters
    ----------

    assignments: mixed array_like of dim (n_run, n_traj, n_cycle) type int
        Assignment of microstates to macrostate labels, where N_runs
        is the number of runs, N_traj is the number of trajectories,
        and N_cycle is the number of cycles.

    transitions: list of list of tuples of ints (run_idx, traj_idx, cycle_idx)
        List of run traces corresponding to transitions. Only the
        first and last trace elements are used.

    weights: mixed array_like of dim (n_run, n_traj, n_cycle) type float or None
        Same shape as the assignments, but gives the weight of the walker at this time.
          (Optional)

    Returns
    -------
    transition_counts : dict of (int, int): int
        Dictionary mapping transitions between macrostate labels to
        the count of microstate transitions between them. If weights
        are given this is the weighted counts.

    """

    # for each transition (start, end) (don't permute) and count them
    # up in a dictionary
    countsmat_d = defaultdict(int)
    for transition in transitions:

        start = transition[0]
        end = transition[-1]

        # get the assignments for the transition
        start_assignment = assignments[start[0]][start[1]][start[2]]
        end_assignment = assignments[end[0]][end[1]][end[2]]


        if weights is not None:
            # get the weight for this walker
            weight = weights[start[0]][start[1]][start[2]]

        else:
            weight = 1.0

        countsmat_d[(start_assignment, end_assignment)] += weight

    return countsmat_d

def counts_d_to_matrix(counts_d):
    """Convert a dictionary of counts for macrostate transitions to an
    assymetric transitions counts matrix.

    Parameters
    ----------
    counts_d : dict of (int, int): int
        Dictionary mapping transitions between macrostate labels to
        the count of microstate transitions between them.

    Returns
    -------
    counts_matrix : numpy.ndarray of int
        Assymetric counts matrix of dim (n_macrostates, n_macrostates).

    """
    # get the number of unique nodes in the counts_d
    max_assignment = max(it.chain(*counts_d.keys()))
    countsmat = np.zeros((max_assignment+1, max_assignment+1))
    for transition, n_trans in counts_d.items():
        countsmat[transition] = n_trans

    return countsmat

def normalize_counts(transition_counts_matrix):
    """Normalize the macrostate outgoing transition counts to 1.0 for each macrostate.

    Parameters
    ----------
    transition_counts_matrix : numpy.ndarray of int
        Assymetric counts matrix of dim (n_macrostates, n_macrostates).

    Returns
    -------
    transition_probability_matrix : numpy.ndarray of float
        Assymetric transition probability matrix of dim (n_macrostates, n_macrostates).

    """

    # if there are any columns that sum to zero we need to set the
    # diagonal value to 1.
    col_sums = transition_counts_matrix.sum(axis=0)

    zero_cols = np.where(col_sums == 0.0)[0]

    if len(zero_cols) > 0:
        transition_counts_matrix = transition_counts_matrix.copy()

        for zero_col_idx in zero_cols:
            transition_counts_matrix[zero_col_idx, zero_col_idx] = 1.0

    return np.divide(transition_counts_matrix, transition_counts_matrix.sum(axis=0))

def transition_counts_matrix(assignments, transitions):
    """Make an asymmetric array of the count of microstate transitions between macrostates.

    Parameters
    ----------
    assignments: mixed array_like of dim (n_run, n_traj, n_cycle) type int
        Assignment of microstates to macrostate labels, where N_runs
        is the number of runs, N_traj is the number of trajectories,
        and N_cycle is the number of cycles.
    transitions: list of list of tuples of ints (run_idx, traj_idx, cycle_idx)
        List of run traces corresponding to transitions. Only the
        first and last trace elements are used.

    Returns
    -------
    transition_counts_matrix : numpy.ndarray of int
        Assymetric counts matrix of dim (n_macrostates, n_macrostates).
    """

    # count them and return as a dictionary
    countsmat_d = transition_counts(assignments, transitions)

    # convert to matrix
    countsmat = counts_d_to_matrix(countsmat_d)

    return countsmat

def transition_probability_matrix(assignments, transitions):
    """Make an asymmetric array of macrostates transition probabilities from microstate assignments.

    Parameters
    ----------
    assignments: mixed array_like of dim (n_run, n_traj, n_cycle) type int
        Assignment of microstates to macrostate labels, where N_runs
        is the number of runs, N_traj is the number of trajectories,
        and N_cycle is the number of cycles.
    transitions: list of list of tuples of ints (run_idx, traj_idx, cycle_idx)
        List of run traces corresponding to transitions. Only the
        first and last trace elements are used.

    Returns
    -------
    transition_probability_matrix : numpy.ndarray of float
        Assymetric transition probability matrix of dim (n_macrostates, n_macrostates).
    """

    # get the counts
    trans_counts_mat = transition_counts_matrix(assignments, transitions)

    # normalize to get the transition probabilities
    trans_prob_mat = normalize_counts(trans_counts_mat)

    return trans_prob_mat

def run_transition_counts_matrix(wepy_hdf5, run_idx, assignment_key, transitions):
    """Generates an asymmetric transition counts matrix directly from a single WepyHDF5 run.

    Parameters
    ----------
    wepy_hdf5 : WepyHDF5 object

    run_idx : int

    assignment_key : str
        Field from trajectory data to use as macrostate label.
    transitions : list of list of tuples of ints (run_idx, traj_idx, cycle_idx)
        List of run traces corresponding to transitions. Only the
        first and last trace elements are used.

    Returns
    -------
    transition_counts_matrix : numpy.ndarray of int
        Assymetric counts matrix of dim (n_macrostates, n_macrostates).

    """


    total_counts_d = defaultdict(int)

    max_assignment = 0
    for transition in transitions:

        start = transition[0]
        end = transition[-1]

        # Gets cluster pair from the hdf5 file
        assignments = wepy_hdf5.get_run_trace_fields(run_idx, [start, end],
                                                  [assignment_key])[assignment_key]

        # Add a count to the cluster pair in the dictionary
        total_counts_d[(assignments[0], assignments[1])] += 1

        # If the assignment index is higher than previously seen
        # assignments, update the max_assignment
        max_assg = max(assignments)
        if max_assg > max_assignment:
            max_assignment = max_assg

    # make a matrix of the counts
    counts_matrix = np.zeros((max_assignment+1, max_assignment+1))
    for transition, n_trans in total_counts_d.items():
            counts_matrix[transition] = n_trans

    return counts_matrix

def run_transition_probability_matrix(wepy_hdf5, run_idx, assignment_key, transitions):
    """Generates an asymmetric transition counts matrix directly from a single WepyHDF5 run.

    Parameters
    ----------
    wepy_hdf5 : WepyHDF5 object

    run_idx : int

    assignment_key : str
        Field from trajectory data to use as macrostate label.
    transitions : list of list of tuples of ints (run_idx, traj_idx, cycle_idx)
        List of run traces corresponding to transitions. Only the
        first and last trace elements are used.

    Returns
    -------
    transition_probability_matrix : numpy.ndarray of float
        Assymetric transition probability matrix of dim (n_macrostates, n_macrostates).

    """

    # get the counts for the run
    counts_mat = run_transition_counts_matrix(wepy_hdf5, run_idx, assignment_key, transitions)

    # normalize to get the probabilities
    trans_prob_matrix = normalize_counts(counts_mat)

    return trans_prob_matrix



from invoke import task

import os
import os.path as osp
from pathlib import Path


def tangle_orgfile(cx, file_path):
    """Tangle the target file using emacs in batch mode. Implicitly dumps
    things relative to the file."""

    cx.run(f"emacs -Q --batch -l org {file_path} -f org-babel-tangle")

@task
def init(cx):
    cx.run("mkdir -p _tangle_source")
    cx.run("mkdir -p _output")

@task
def clean(cx):
    cx.run("rm -rf _tangle_source")
    cx.run("rm -rf _output")

@task(pre=[init])
def tangle(cx):
    tangle_orgfile(cx, "README.org")


@task
def clean_env(cx):
    cx.run("rm -rf _env")

@task(pre=[init])
def env(cx):
    """Create the environment from the specs in 'env'. Must have the
    entire repository available as it uses the tooling from it.

    """

    example_name = Path(os.getcwd()).stem

    with cx.cd("../../../"):
        cx.run(f"inv docs.env-tutorial -n {example_name}")



"""Module for generating wepy systems"""

from copy import copy, deepcopy

import numpy as np

import simtk.openmm.app as omma
import simtk.openmm as omm
import simtk.unit as unit

import mdtraj as mdj

from wepy.util.mdtraj import mdtraj_to_json_topology
from wepy.util.json_top import (json_top_residue_fields,
                                json_top_residue_df,
                                json_top_atom_df,
                                json_top_subset)


# OpenMM helpers
from wepy.runners.openmm import OpenMMRunner, OpenMMState
from wepy.walker import Walker
from wepy.runners.openmm import gen_walker_state, UNIT_NAMES, GET_STATE_KWARG_DEFAULTS
from wepy.orchestration.snapshot import WepySimApparatus, SimSnapshot
from wepy.orchestration.configuration import Configuration
from wepy.orchestration.orchestrator import Orchestrator

# resamplers
from wepy.resampling.resamplers.resampler import NoResampler
from wepy.resampling.resamplers.wexplore import WExploreResampler
from wepy.resampling.resamplers.revo import REVOResampler

# boundary conditions
from wepy.boundary_conditions.boundary import NoBC, RandomBC

# integrators
from simtk.openmm import LangevinIntegrator

# mappers
from wepy.work_mapper.mapper import Mapper
from wepy.work_mapper.worker import WorkerMapper, Worker
from wepy.work_mapper.task_mapper import TaskMapper, WalkerTaskProcess

from wepy.runners.openmm import (
    OpenMMCPUWorker, OpenMMGPUWorker,
    OpenMMCPUWalkerTaskProcess, OpenMMGPUWalkerTaskProcess,
)

# reporters
from wepy.reporter.hdf5 import WepyHDF5Reporter

from wepy.reporter.dashboard import (
    DashboardReporter,
    ResamplerDashboardSection,
    RunnerDashboardSection,
    BCDashboardSection)
from wepy.reporter.wexplore.dashboard import WExploreDashboardSection
from wepy.reporter.revo.dashboard import REVODashboardSection
from wepy.reporter.openmm import OpenMMRunnerDashboardSection

from wepy.reporter.restree import ResTreeReporter
from wepy.reporter.walker import WalkerReporter


# extras for reporters
from wepy.runners.openmm import UNIT_NAMES

# workers


class OpenMMSimMaker():

    ### Default settings

    ## Resamplers

    RESAMPLERS = [NoResampler, WExploreResampler, REVOResampler,]

    WEXPLORE_DEFAULTS = {
        'pmax' : 0.5,
        'pmin' : 1e-12,
        'max_n_regions' : (10, 10, 10, 10),
        # ALERT: shouldn't be automatically setting these distances
        # 'max_region_sizes' : (1, 0.5, 0.35, 0.25),
    }


    REVO_DEFAULTS = {
        'pmax' : 0.5,
        'pmin' : 1e-12,
        'char_dist' : 1,
        # ALERT: shouldn't be automatically setting these distances
        # 'merge_dist' : 2.5,
        'dist_exponent' : 4,
        'weights' : True,
    }

    DEFAULT_RESAMPLER_PARAMS = {
        'NoResampler' : {},
        'WExploreResampler' : WEXPLORE_DEFAULTS,
        'REVOResampler' : REVO_DEFAULTS,
    }



    ## BCs

    BCS = [NoBC, RandomBC,]

    DEFAULT_BC_PARAMS = {
        'NoBC' : {},
        'RandomBC' : {},
    }


    ## Integrators

    INTEGRATORS = [LangevinIntegrator,]

    #(TEMPERATURE, FRICTION_COEFFICIENT, STEP_SIZE)
    LANGEVIN_DEFAULTS = (
        300.0*unit.kelvin,
        1/unit.picosecond,
        0.002*unit.picoseconds,
    )

    DEFAULT_INTEGRATOR_PARAMS = {
        'LangevinIntegrator' : LANGEVIN_DEFAULTS,
    }

    # these are used just to generate states
    INTEGRATOR_FIXTURE = omm.LangevinIntegrator
    INTEGRATOR_FIXTURE_PARAMS = DEFAULT_INTEGRATOR_PARAMS['LangevinIntegrator']

    ## OpenMM platforms

    DEFAULT_PLATFORM_PARAMS = {
        'Reference' : {},
        'CPU' : {},
        'OpenCL' : {},
        'CUDA' : {},

    }

    ## other runner options
    GET_STATE_KWARGS = {}

    ## Work Mappers
    MAPPERS = [Mapper, WorkerMapper, TaskMapper]

    DEFAULT_MAPPER_PARAMS = {
        'Mapper' : {},
        'WorkerMapper' : {},
        'TaskMapper' : {},
    }


    ## Reporters
    REPORTERS = [
        WepyHDF5Reporter,
        DashboardReporter,
        ResTreeReporter,
        WalkerReporter,
    ]

    WEPY_HDF5_REPORTER_DEFAULTS = {
                        'main_rep_idxs' : None,
                        'save_fields' : None,
                        'units' : dict(UNIT_NAMES),
                        'sparse_fields' : {'velocities' : 10},
                        'all_atoms_rep_freqs' : 10,
                        'alt_reps' : None,
                        'swmr_mode' : True,
                    }

    RESTREE_REPORTER_DEFAULTS = {
        'node_radius' : 3.0,
        'row_spacing' : 5.0,
        'step_spacing' : 20.0,
        'colormap_name' : 'plasma',
    }

    DEFAULT_REPORTER_PARAMS = {
        'WepyHDF5Reporter' : WEPY_HDF5_REPORTER_DEFAULTS,
        'DashboardReporter' : {},
        'ResTreeReporter' : RESTREE_REPORTER_DEFAULTS,
        'WalkerReporter' : {},
    }


    def __init__(self,
                 distance=None,
                 init_state=None,
                 system=None,
                 topology=None,
    ):

        self.distance = distance
        self.init_state = init_state
        self.system = system
        self.topology = topology

        self.getState_kwargs = dict(GET_STATE_KWARG_DEFAULTS)
        if self.GET_STATE_KWARGS is not None:
            self.getState_kwargs.update(self.GET_STATE_KWARGS)

    def make_state(self, system, positions):

        # a temporary integrator just for this
        integrator = self.INTEGRATOR_FIXTURE(*self.INTEGRATOR_FIXTURE_PARAMS)

        init_state = gen_walker_state(positions, system, integrator,
                                      getState_kwargs=self.getState_kwargs)

        return init_state

    @classmethod
    def make_initial_walkers(cls, state, n_walkers):

        init_weight = 1.0 / n_walkers
        init_walkers = [Walker(deepcopy(state), init_weight) for i in range(n_walkers)]

        return init_walkers

    def make_apparatus(self,
                       platform='Reference',
                       platform_params=None,
                       runner_params=None,
                       integrator='LangevinIntegrator',
                       integrator_params=None,
                       resampler='WExploreResampler',
                       resampler_params=None,
                       bc=None,
                       bc_params=None,
    ):

        ## RUNNER

        # choose which integrator to use
        integrator_class = [i for i in self.INTEGRATORS
                            if i.__name__ == integrator][0]

        integrator_name = integrator_class.__name__


        # use either the default params or the user params
        if integrator_params is None:
            integrator_params = self.DEFAULT_INTEGRATOR_PARAMS[integrator_name]

        integrator = integrator_class(*integrator_params)

        # TODO: not handling the params here
        if platform_params is None:
            platform_params = self.DEFAULT_PLATFORM_PARAMS[platform]

        # handle additional runner options
        if runner_params is None:
            runner_params = {}

        # make the runner for the test system
        runner = OpenMMRunner(self.system,
                              self.topology,
                              integrator,
                              platform=platform,
                              **runner_params)


        # RESAMPLER

        # choose which resampler to use
        resampler_class = [res for res in self.RESAMPLERS
                           if res.__name__ == resampler][0]

        resampler_name = resampler_class.__name__

        # use either the default params or the user params
        if resampler_params is None:
            resampler_params = self.DEFAULT_RESAMPLER_PARAMS[resampler_name]


        resampler = resampler_class(distance=self.distance,
                                   init_state=self.init_state,
                                   **resampler_params)


        # BOUNDARY_CONDITIONS

        # you don't have to specify a boundary condition
        bc_name = bc
        if bc_name is not None:


            # choose which bc to use
            bc_class = [res for res in self.BCS
                        if res.__name__ == bc_name][0]

            bc_name = bc_class.__name__

            # use either the default params or the user params
            if bc_params is None:
                bc_params = self.DEFAULT_BC_PARAMS[bc_name]

            bc = self.make_bc(bc_class, bc_params)




        # APPARATUS

        # build the apparatus
        sim_apparatus = WepySimApparatus(runner, resampler=resampler,
                                         boundary_conditions=bc)

        return sim_apparatus


    def make_bc(self, bc_class, bc_params):

        bc = bc_class(**bc_params)

        return bc


    @classmethod
    def choose_work_mapper_platform_params(cls, platform, mapper_name):

        work_mapper_params = {}

        if mapper_name == 'WorkerMapper':

            if platform == 'Reference':
                worker_type = Worker
            elif platform == 'CPU':
                worker_type = OpenMMCPUWorker
            elif platform in ('CUDA', 'OpenCL',):
                worker_type = OpenMMGPUWorker
            else:
                worker_type = Worker

            work_mapper_params['worker_type'] = worker_type

        elif mapper_name == 'TaskMapper':
            if platform == 'Reference':
                worker_type = Worker
            elif platform == 'CPU':
                worker_type = OpenMMCPUWalkerTaskProcess
            elif platform in ('CUDA', 'OpenCL',):
                worker_type = OpenMMGPUWalkerTaskProcess
            else:
                worker_type = Worker

            work_mapper_params['walker_task_type'] = worker_type

        return work_mapper_params

    def choose_dashboard_sections(self, apparatus):

        # defaults for the dashboard sections
        if apparatus.boundary_conditions is not None:
            dashboard_sections = {'resampler' : ResamplerDashboardSection(apparatus.resampler),
                                  'runner' : RunnerDashboardSection(apparatus.runner),
                                  'bc' : BCDashboardSection(apparatus.boundary_conditions),
            }

        else:
            dashboard_sections = {'resampler' : ResamplerDashboardSection(apparatus.resampler),
                                  'runner' : RunnerDashboardSection(apparatus.runner),
                                  'bc' : None,
            }

        # the choices here should be fairly portable between different
        # systems as they are not specific to any system details

        ## resampler

        # WExplore
        if type(apparatus.resampler).__name__ == 'WExploreResampler':
            dashboard_sections['resampler'] = WExploreDashboardSection(apparatus.resampler)
        # REVO
        elif type(apparatus.resampler).__name__ == 'REVOResampler':
            dashboard_sections['resampler'] = REVODashboardSection(apparatus.resampler)

        ## BC
        # NoBC
        if type(apparatus.boundary_conditions).__name__ == 'NoBC':
            dashboard_sections['bc'] = BCDashboardSection(apparatus.boundary_conditions)
        # Random
        elif type(apparatus.boundary_conditions).__name__ == 'RandomBC':
            dashboard_sections['bc'] = BCDashboardSection(apparatus.boundary_conditions)

        ## Runner

        # OpenMM
        if type(apparatus.runner).__name__ == 'OpenMMRunner':
            dashboard_sections['runner'] = OpenMMRunnerDashboardSection(
                apparatus.runner)

        return dashboard_sections


    def resolve_reporter_params(self, apparatus, reporter_specs, reporters_kwargs=None):

        if reporters_kwargs is not None:
            raise NotImplementedError("Only the defaults are supported currently")

        # ellipsis means use all of the defaults
        if reporter_specs is Ellipsis:
            # defaults to use
            reporter_specs = [
                'WepyHDF5Reporter',
                'DashboardReporter',
                # DEBUG: this isn't compatible right now, needs refactoring
                #'ResTreeReporter',
                'WalkerReporter',
            ]

        # if it is None, we use no reporters
        elif reporter_specs is None:
            return [], []

        # augment the dashboard with the sections relevant to our
        # components
        if 'DashboardReporter' in reporter_specs:

            # customize the dashboard based on what components are in
            # the apparatus or what a subclass has customized this for
            dashboard_sections = self.choose_dashboard_sections(apparatus)

        # get the actual classes
        reporter_classes = []
        for reporter_spec in reporter_specs:
            match = False
            for reporter_class in self.REPORTERS:

                if reporter_class.__name__ == reporter_spec:
                    match = reporter_class
                    break

            if not match:
                raise ValueError("Unkown reporter for spec {}".format(reporter_spec))

            reporter_classes.append(match)


        # then get the default params for them and commensurate them
        # with the given kwargs
        reporters_params = []
        for reporter_spec in reporter_specs:
            reporter_params = {}

            if reporter_spec == 'WepyHDF5Reporter':

                # always set these ones automatically
                auto_params = {
                    'topology' : self.json_top(),
                    'resampler' : apparatus.resampler,
                    'boundary_conditions' : apparatus.boundary_conditions,
                }

                reporter_params.update(auto_params)
                reporter_params.update(deepcopy(self.DEFAULT_REPORTER_PARAMS[reporter_spec]))

            elif reporter_spec == 'DashboardReporter':

                # always set these ones automatically
                auto_params = {
                    # the 'getStepSize' method is an abstract one for
                    # all integrators so we can rely on it being here.
                    'step_time' : apparatus.runner.integrator.getStepSize(),

                    # the dashboard sections
                    'resampler_dash' : dashboard_sections['resampler'],
                    'runner_dash' : dashboard_sections['runner'],
                    'bc_dash' : dashboard_sections['bc'],
                }

                reporter_params.update(auto_params)
                reporter_params.update(deepcopy(self.DEFAULT_REPORTER_PARAMS[reporter_spec]))

            elif reporter_spec == 'WalkerReporter':

                # always set these ones automatically
                auto_params = {
                    'json_topology' : self.json_top(),
                    'init_state' : self.init_state,
                }

                reporter_params.update(auto_params)
                reporter_params.update(deepcopy(self.DEFAULT_REPORTER_PARAMS[reporter_spec]))

            elif reporter_spec == 'ResTreeReporter':

                # always set these ones automatically
                auto_params = {
                    'resampler' : apparatus.resampler,
                    'boundary_condition' : apparatus.boundary_conditions,
                }

                reporter_params.update(auto_params)
                reporter_params.update(deepcopy(self.DEFAULT_REPORTER_PARAMS[reporter_spec]))

            else:
                reporter_params.update(deepcopy(self.DEFAULT_REPORTER_PARAMS[reporter_spec]))

            # add them to the list for this reporter
            reporters_params.append(reporter_params)

        return reporter_classes, reporters_params


    def make_configuration(self,
                           apparatus,
                           work_mapper_class=None,
                           work_mapper_spec='TaskMapper',
                           work_mapper_params=None,
                           platform='Reference',
                           # defaults to using all of the defaults
                           reporters=Ellipsis,
                           reporter_kwargs=None,
                           work_dir=None,
                           monitor_class=None,
                           monitor_params=None,
    ):

        # MAPPER

        # choose which mapper to use

        # use the class if given
        if work_mapper_class is not None:
            pass

        # use the spec string given
        elif work_mapper_spec is not None:

            work_mapper_class = [mapper for mapper in self.MAPPERS
                                 if mapper.__name__ == work_mapper_spec][0]

        else:
            raise ValueError("neither work_mapper_class or work_mapper_spec were not given")

        mapper_name = work_mapper_class.__name__

        # use either the default params or the user params
        if work_mapper_params is None:
            work_mapper_params = self.DEFAULT_MAPPER_PARAMS[mapper_name]

        # depending on the platform and work mapper choose the worker
        # type and update the params in place
        work_mapper_params.update(
            self.choose_work_mapper_platform_params(platform, mapper_name))


        # REPORTERS

        reporter_classes, reporter_params = \
                        self.resolve_reporter_params(apparatus, reporters, reporter_kwargs)


        ## Monitor


        config = Configuration(
            work_mapper_class=work_mapper_class,
            work_mapper_partial_kwargs=work_mapper_params,
            reporter_classes=reporter_classes,
            reporter_partial_kwargs=reporter_params,
            work_dir=work_dir,
            monitor_class=monitor_class,
            monitor_partial_kwargs=monitor_params,
        )

        return config


    def make_sim_manager(self, n_walkers, apparatus, config):

        walkers = self.make_initial_walkers(self.init_state, n_walkers)
        snapshot = SimSnapshot(walkers, apparatus)

        sim_manager = Orchestrator.gen_sim_manager(snapshot, config)

        return sim_manager



class OpenMMToolsTestSysSimMaker(OpenMMSimMaker):

    TEST_SYS = None

    @classmethod
    def num_atoms(cls):

        json_top = cls.json_top()

        # get the atom dataframe and select them from the ligand residue
        return len(json_top_atom_df(json_top))

    @classmethod
    def box_vectors(cls):

        # just munge the thing they give you to be a nice array Quantity
        bvs = cls.TEST_SYS().system.getDefaultPeriodicBoxVectors()

        return np.array([bv.value_in_unit(unit.nanometer) for bv in bvs]) * unit.nanometer

    @classmethod
    def json_top(cls):

        test_sys = cls.TEST_SYS()

        # convert to a JSON top
        json_top = mdtraj_to_json_topology(mdj.Topology.from_openmm(test_sys.topology))

        return json_top



"""Distance metrics related to common features of receptor-ligand
processes in molecular systems.

The ReceptorDistance class is an abstract class that provides some
common functionality for normalizing reference states, providing
correct indices of receptor and ligand atoms, and a common image
function.

Subclasses of ReceptorDistance need only implement the
'image_distance' function according to their needs.

The UnbindingDistance is a useful metric for enhancing ligand movement
away from the reference bound state conformation.

The RebindingDistance is a useful metric for enhancing the movement of
a ligand towards a reference state.

"""

import logging

import numpy as np

from wepy.util.util import box_vectors_to_lengths_angles

from geomm.grouping import group_pair
from geomm.superimpose import superimpose
from geomm.rmsd import calc_rmsd
from geomm.centering import center_around

from wepy.resampling.distances.distance import Distance

class ReceptorDistance(Distance):
    """Common abstract class for receptor-ligand molecular systems.

    Any input walker state should have the attributes 'positions' and
    'box_vectors' and will first preprocess these states, by
    recentering the complex. Two component systems like
    receptor-ligands in periodic boxes can sometimes drift to
    different periodic images of the box making them unsuitable for
    RMSD calculations.

    The 'image' method will align the receptor structures of the state
    to the reference state after the recentering.

    """

    def __init__(self, ligand_idxs, binding_site_idxs, ref_state):
        """Construct a distance metric.

        Parameters
        ----------

        ligand_idxs : arraylike of int
            The indices of the atoms from the 'positions' attribute of
            states that correspond to the ligand molecule.

        binding_site_idxs : arraylike of int
            The indices of the atoms from the 'positions' attribute of
            states that correspond to the atoms of the binding site of
            the receptor. These are the atoms you want to perform
            alignment on and so can be a subset of a complete molecule.

        ref_state : object implementing WalkerState
            The reference state all walker states will be aligned to
            with 'positions' (Nx3 dims) and 'box_vectors' (3x3 array)
            attributes.

        """


        # the idxs of the ligand and binding site from the whole state
        self._lig_idxs = ligand_idxs
        self._bs_idxs = binding_site_idxs

        # number of atoms in each
        self._n_lig_atoms = len(self._lig_idxs)
        self._n_bs_atoms = len(self._bs_idxs)

        # the idxs used for the whole image
        self._image_idxs = np.concatenate( (self._lig_idxs, self._bs_idxs) )

        # the idxs of the ligand and binding site within the image
        self._image_lig_idxs = np.arange(self._n_lig_atoms)
        self._image_bs_idxs = np.arange(self._n_lig_atoms, self._n_lig_atoms + self._n_bs_atoms)

        # save the reference state's image so we can align all further
        # images to it

        self.ref_image = self._unaligned_image(ref_state)

    def _unaligned_image(self, state):
        """The preprocessing method of states.

        First it groups the binding site and ligand into the same
        periodic box image and then centers the box around their
        mutual center of mass and returns only the positions of the
        binding site and ligand.

        Parameters
        ----------
        state : object implementing WalkerState
            State with 'positions' (Nx3 dims) and 'box_vectors' (3x3
            array) attributes.

        Returns
        -------

        """

        # get the box lengths from the vectors
        box_lengths, box_angles = box_vectors_to_lengths_angles(state['box_vectors'])

        # recenter the protein-ligand complex into the center of the
        # periodic boundary conditions

        # regroup the ligand and protein in together
        grouped_positions = group_pair(state['positions'], box_lengths,
                                    self._bs_idxs, self._lig_idxs)

        # then center them around the binding site
        centered_positions = center_around(grouped_positions, self._bs_idxs)

        # slice these positions to get the image
        state_image = centered_positions[self._image_idxs]

        return state_image

    def image(self, state):
        """Transform a state to a receptor image.

        A receptor image is one in which the binding site and ligand
        are first normalized to be within the same periodic box image
        and at the center of the box, and then only the binding site
        and ligand are retained.

        Parameters
        ----------
        state : object implementing WalkerState
            State with 'positions' (Nx3 dims) and 'box_vectors' (3x3 array)
            attributes.

        Returns
        -------

        receptor_image : array of float
            The positions of binding site and ligand after
            preprocessing.

        """

        # get the unaligned image
        state_image = self._unaligned_image(state)

        # then superimpose it to the reference structure
        sup_image, _, _ = superimpose(self.ref_image, state_image, idxs=self._image_bs_idxs)

        return sup_image


class UnbindingDistance(ReceptorDistance):
    """Distance metric for measuring differences between walker states in
    regards to the RMSDs between ligands.

    Images are produced using the ReceptorDistance.image method. The
    distance between images then is solely the RMSD of the ligand
    atoms.

    Thus walkers were ligands are divergent in position will have
    higher distances.

    """

    def image_distance(self, image_a, image_b):

        # we calculate the rmsd of only the ligands between the
        # images
        lig_rmsd = calc_rmsd(image_a, image_b, idxs=self._image_lig_idxs)

        return lig_rmsd

class RebindingDistance(ReceptorDistance):
    """Distance metric for measuring differences between walker states in
    regards to the RMSDs between ligands.

    Images are produced using the ReceptorDistance.image method. The
    distance between images then is the relative difference between
    the ligand RMSDs to the reference state.

    """

    def image_distance(self, image_a, image_b):

        # we calculate the rmsd of only the ligands between each image
        # and the reference
        state_a_rmsd = calc_rmsd(self.ref_image, image_a, idxs=self._image_lig_idxs)
        state_b_rmsd = calc_rmsd(self.ref_image, image_b, idxs=self._image_lig_idxs)

        # then we get the absolute value of the reciprocals of these rmsd
        # values
        d = abs(1./state_a_rmsd - 1./state_b_rmsd)

        return d



"""Library of and framework for making boundary conditions in wepy simulations.

In many simulations you would like to apply external constraints upon
the dynamics of a simulation. For example in ligand unbinding
simulations you may want to recognize and act on the fact that a
ligand has reached some sort of threshold distance away from it's
originally bound protein. The `LigandUnbinding` boundary condition
class does exactly this and modifies the walker state to restart it in
the original starting position instead of continuing the dynamics of
the runner. Another possible way to implement this would be to only
change a single data field in the walker state that tells you the
"color" of the walker which indicates the last boundary it crossed,
without interrupting the dynamical continuity of the
simulation. However, the analysis routines in `wepy` are adapted to
deal with such discontinuities if you desire such non-equilibrium
simulations.

Use of boundary conditions is so common that the `wepy` simulation
manager and `WepyHDF5` storage backend provide an interface for
applying and saving results of boundary conditions.

Boundary conditions need not be used in a simulation at all however
and are strictly optional.

This module provides a small library of common boundary conditions and
a reference for the implementation of your own boundary
conditions. The approach to have a microframework for making boundary
conditions was chosen because the actual details of them are highly
domain dependent and it is unlikely that `wepy` could begin to cover
all of them.

A boundary condition is simply a duck typed object that follows the
interface called by the simulation manager. This means that you need
not use this microframework at all and can implement one from the
ground up. I.e. no explicit class inheritance checks are done at any
point.

However, to make things easy just inherit from the `BoundaryCondition`
base class and override the necessary functions.

These methods must be implemented to work correctly in the reference
simulation manager:
- warp_walkers
- progress
- update_bc


A non-trivial boundary condition (BC for short) will override at least
the `warp_walkers` method. "Warping" in `wepy` terminology means for a
walker to have met the constraints of the BC and have it's state be
modified in response (think warping through space via a wormhole like
in star trek). Warps can be 'continuous' or 'discontinuous' depending
on whether the continuity of dynamics is broken due to this
effect. For example, resetting the positions of a molecular dynamics
to its starting point is discontinuous, and the change of a string
label or some other field not used by the runner to compute dynamics
is continuous.

For each warping event a single warping record is generated, that
should detail the event.

The simulation manager will also make calls to `progress` and
`update_bc`, but these are accessory to the `warp_walkers` method.

The `progress` call returns a continual record (one per cycle) and is
intended to report on some metrics of progress the walkers have
achieved. For example in the ligand unbinding example this can be the
minimum distance to the protein. The progress record can contain
multiple features per walker. It can also be used to make economical
use of calculations used in the BC when determining the warpings of
the walkers.

Additionally for routines in `wepy.analysis` to work properly the
`warping_discontinuity` classmethod (can be called from the class
without an object instance) must be implemented correctly.

The `update_bc` call is used to get changes to the BC object
itself. To use the ligand unbinding example one might lower the
threshold for unbinding if so many cycles go by without an event
occuring with perhaps introspection on the progress values.

Any changes in the state of the BC object should be happen in the call
stack of `update_bc` and reflected in the 'bc_record' returned to the
simulation manager.

Record groups generated by a BC class are:
- warping
- boundary_conditions
- progress

It is recommended (but optional) to have all the field names, shapes,
and dtypes for the records in the class definition for these
where possible.

Notes
-----

Boundary conditions in wepy simulations are optional.

Inherit from the
`wepy.boundary_conditions.boundary.BoundaryConditions` class to get
started developing your own boundary conditions.

Boundary conditions must implement methods: `warp_walkers`,
`progress`, and `update_bc` for simulation.

It is the programmers responsibility to handle discontinuous warping
events by implementing the `warping_discontinuity` class method.


References
----------

Examples
--------

"""



import numpy as np

# this is a typical function for mapping over the dataset. I have
# added the extra argument scaling_factor just to demonstrate that you
# can have other arguments.
def traj_field_lj_dist(traj_data, scaling_factor):

    positions = traj_data['positions']

    # slice out positions for each LJ particle
    lj1 = positions[:,0,:]
    lj2 = positions[:,1,:]
    # compute distances with the scaling factor
    distances = scaling_factor * np.sqrt(
        (lj1[:,0] - lj2[:,0])**2 + (lj1[:,1] - lj2[:,1])**2 + (lj1[:,2] - lj2[:,2])**2)

    return distances



if __name__ == "__main__":
    from wepy.hdf5 import WepyHDF5

    # load the HDF5 file in read/write so we can save data to the
    # observables
    wepy_hdf5_path = "_output/we/results.wepy.h5"
    wepy_h5 = WepyHDF5(wepy_hdf5_path, mode='r+')

    with wepy_h5:
        wepy_h5.compute_observable(traj_field_lj_dist, ['positions'],
                                   (2.0,),
                                   save_to_hdf5='2*rmsd',
                                   map_func=map)

        wepy_h5.compute_observable(traj_field_lj_dist, ['positions'],
                                   (1.0,),
                                   save_to_hdf5='rmsd',
                                   map_func=map)



"""Modular component for defining distance metrics usable within many
different resamplers.

This module contains an abstract base class for Distance classes.

The suggested implementation method is to leave the 'distance' method
as is, and override the 'image' and 'image_distance' methods
instead. Because the default 'distance' method calls these
transparently. The resamplers will use the 'image' and
'image_distance' calls because this allows performance optimizations.

For example in WExplore the images for some walkers end up being
stored as the definitions of Voronoi regions, and if the whole walker
state was stored it would not only use much more space in memory but
require that common transformations be repeated every time a distance
is to be calculated to that image (which is many times). In REVO all
to all distances between walkers are computed which would also incur a
high cost.

So use the 'image' to do precomputations on raw walker states and use
the 'image_distance' to compute distances using only those images.

"""
import logging
import numpy as np

class Distance(object):
    """Abstract Base class for Distance classes."""

    def __init__(self):
        """Constructor for Distance class."""
        pass

    def image(self, state):
        """Compute the 'image' of a walker state which should be some
        transformation of the walker state that is more
        convenient. E.g. for precomputation of expensive operations or
        for saving as resampler state.

        The abstract implementation is naive and just returns the
        state itself, thus it is the identity function.

        Parameters
        ----------
        state : object implementing WalkerState
            The state which will be transformed to an image

        Returns
        -------
        image : object implementing WalkerState
            The same state that was given as an argument.

        """

        return state

    def image_distance(self, image_a, image_b):
        """Compute the distance between two images of walker states.

        Parameters
        ----------
        image_a : object produced by Distance.image

        image_b : object produced by Distance.image

        Returns
        -------

        distance : float
            The distance between the two images

        Raises
        ------

        NotImplementedError : always because this is abstract

        """
        raise NotImplementedError

    def distance(self, state_a, state_b):
        """Compute the distance between two states.

        Parameters
        ----------
        state_a : object implementing WalkerState

        state_b : object implementing WalkerState

        Returns
        -------

        distance : float
            The distance between the two walker states


        """

        return self.image_distance(self.image(state_a),
                                      self.image(state_b))



class XYEuclideanDistance(Distance):
    """2 dimensional euclidean distance between points. 

    States have the attributes 'x' and 'y'.

    """

    def image(self, state):

        return np.array([state['x'], state['y']])


    def image_distance(self, image_a, image_b):

        return np.sqrt((image_a[0] - image_b[0])**2 +
                       (image_a[1] - image_b[1])**2)



"""Very basic example of a simulation without a resampler or boundary
conditions"""

from copy import copy

import simtk.openmm as omm
import simtk.unit as unit

from openmm_systems.test_systems import LennardJonesPair

from wepy.resampling.resamplers.resampler import NoResampler
from wepy.runners.openmm import OpenMMRunner, gen_walker_state
from wepy.walker import Walker
from wepy.sim_manager import Manager

# use a ready made system for OpenMM MD simulation
test_sys = LennardJonesPair()

integrator = omm.LangevinIntegrator(300.0*unit.kelvin,
                                    1/unit.picosecond,
                                    0.002*unit.picoseconds)

init_state = gen_walker_state(test_sys.positions, test_sys.system, integrator)

runner = OpenMMRunner(test_sys.system, test_sys.topology, integrator,
                      platform='Reference')

# a trivial resampler which does nothing
resampler = NoResampler()

# Run the simulation

# number of cycles of WE to perform
n_cycles = 5

# the number of MD dynamics steps for each cycle
n_steps = 1000
steps = [n_steps for i in range(n_cycles)]

# number of parallel simulations
n_walkers = 10



# create the initial walkers with equal weights
init_weight = 1.0 / n_walkers
init_walkers = [Walker(copy(init_state), init_weight) for i in range(n_walkers)]

sim_manager = Manager(init_walkers,
                      runner=runner,
                      resampler=resampler)

# run the simulation and get the results
final_walkers, _ = sim_manager.run_simulation(n_cycles, steps)



"""Miscellaneous functions used in wepy that don't fit elsewhere.

Warnings
--------

Deprecation warning!

Try not to rely on these functions as part of your stable code. These
were included here in wepy because other libraries do not provide this
functionality, although they probably should. That is these functions
are outside of the purview of the goals of wepy and are not guaranteed
to be included in wepy indefinitely.


Please do not rely on anything in the util module to exist
indefinitely.

Any function is subject to deprecation at any release.

If you use these functions consider copying the source code.

If any of these functions or classes become reified and officially
part of wepy they will go in a sub-module dedicated to their purposes.

But these will likely (and should) be moved to other libraries as they
are more general purpose than wepy.

"""



import itertools as it
from collections import defaultdict
from warnings import warn

import logging
from eliot import start_action, log_call

import numpy as np

from wepy.resampling.decisions.decision import Decision, NoDecision

class ResamplerError(Exception):
    """Error raised when some constraint on resampling properties is
    violated."""
    pass

class Resampler():
    """Abstract base class for implementing resamplers.

    All subclasses of Resampler must implement the 'resample' method.

    If extra reporting on resampling and resampler updates desired
    subclassed resamplers should update the following class constants
    for specifying the decision class and the names, shapes, data
    types, and table-like records for each record group:

    - DECISION
    - RESAMPLING_FIELDS
    - RESAMPLING_SHAPES
    - RESAMPLING_DTYPES
    - RESAMPLING_RECORD_FIELDS
    - RESAMPLER_FIELDS
    - RESAMPLER_SHAPES
    - RESAMPLER_DTYPES
    - RESAMPLER_RECORD_FIELDS

    The DECISION constant should be a
    wepy.resampling.decisions.decision.Decision subclass.

    This base class provides some hidden methods that are useful for
    various purposes.

    To help maintain constraints on the number of walkers in a
    simulation the constructor allows for setting of a minimum and/or
    maximum of walkers.

    These values are allowed to be either integers, None, or Ellipsis.

    Integers set hard values for the minimum and maximum values.

    None indicates that the min and max are unbounded. For the min
    this translates to a value of 1 since there must always be one
    walker.

    Ellipsis is an indicator to determine the minimum and maximum
    dependent on the number of walkers provided for resampling.

    If the max_num_walkers is Ellipsis and the number of walkers given
    is 10 then the maximum will be set to 10 for that
    resampling. Conversely, for if the minimum is Ellipsis.

    If both the min and max are set to Ellipsis then the number of
    walkers is always kept the same.

    Note that this does not implement any algorithm that actually
    decides how many walkers there will be but just checks that these
    constraints are met by those implementations in subclasses.

    To allow for this checking the '_resample_init' method should be
    called at the beginning of the 'resample' method and the
    '_resample_cleanup' should be called at the end of the 'resample'
    method.



    """

    DECISION = Decision
    """The decision class for this resampler."""

    CYCLE_FIELDS = ('step_idx', 'walker_idx',)
    """The fields that get added to the decision record for all resampling
    records. This places a record within a single destructured listing
    of records for a single cycle of resampling using the step and
    walker index.
    """

    CYCLE_SHAPES = ((1,), (1,),)
    """Data shapes of the cycle fields."""

    CYCLE_DTYPES = (np.int, np.int,)
    """Data types of the cycle fields """

    CYCLE_RECORD_FIELDS = ('step_idx', 'walker_idx',)
    """Optional, names of fields to be selected for truncated
    representation of the record group.
    """

    # data for resampling performed (continual)
    RESAMPLING_FIELDS = DECISION.FIELDS + CYCLE_FIELDS
    """String names of fields produced in this record group.

    Resampling records are typically used to report on the details of
    how walkers are resampled for a given resampling step.

    Warning
    -------

    This is a critical function of many other components of the wepy
    framework and probably shouldn't be altered by most developers.

    Thi is where the information about cloning and merging of walkers
    is given. Seeing as this is a most of the value proposition of
    wepy as a tool getting rid of it will render most of the framework
    useless.

    But sticking to the 'loosely coupled, tightly integrated' mantra
    you are free to modify these fields. This would be useful for
    implementing resampling strategies that do not follow basic
    cloning and merging. Just beware, that most of the lineage based
    analysis will be broken without implementing a new Decision class.

    """

    RESAMPLING_SHAPES = DECISION.SHAPES + CYCLE_SHAPES
    """Numpy-style shapes of all fields produced in records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A tuple of ints that specify the shape of the field element
       array.

    B. Ellipsis, indicating that the field is variable length and
       limited to being a rank one array (e.g. (3,) or (1,)).

    C. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    Note that the shapes must be tuple and not simple integers for rank-1
    arrays.

    Option B will result in the special h5py datatype 'vlen' and
    should not be used for large datasets for efficiency reasons.

    """

    RESAMPLING_DTYPES = DECISION.DTYPES + CYCLE_DTYPES
    """Specifies the numpy dtypes to be used for records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A `numpy.dtype` object.

    D. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    """

    RESAMPLING_RECORD_FIELDS = DECISION.RECORD_FIELDS + CYCLE_RECORD_FIELDS
    """Optional, names of fields to be selected for truncated
    representation of the record group.

    These entries should be strings that are previously contained in
    the 'FIELDS' class constant.

    While strictly no constraints on to which fields can be added here
    you should only choose those fields whose features could fit into
    a plaintext csv or similar format.

    """

    # changes to the state of the resampler (sporadic)
    RESAMPLER_FIELDS = ()
    """String names of fields produced in this record group.

    Resampler records are typically used to report on changes in the
    state of the resampler.

    Notes
    -----

    These fields are not critical to the proper functioning of the
    rest of the wepy framework and can be modified freely.

    However, reporters specific to this resampler probably will make
    use of these records.

    """

    RESAMPLER_SHAPES = ()
    """Numpy-style shapes of all fields produced in records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A tuple of ints that specify the shape of the field element
       array.

    B. Ellipsis, indicating that the field is variable length and
       limited to being a rank one array (e.g. (3,) or (1,)).

    C. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    Note that the shapes must be tuple and not simple integers for rank-1
    arrays.

    Option B will result in the special h5py datatype 'vlen' and
    should not be used for large datasets for efficiency reasons.

    """

    RESAMPLER_DTYPES = ()
    """Specifies the numpy dtypes to be used for records.

    There should be the same number of elements as there are in the
    corresponding 'FIELDS' class constant.

    Each entry should either be:

    A. A `numpy.dtype` object.

    D. None, indicating that the first instance of this field will not
       be known until runtime. Any field that is returned by a record
       producing method will automatically interpreted as None if not
       specified here.

    """

    RESAMPLER_RECORD_FIELDS = ()
    """Optional, names of fields to be selected for truncated
    representation of the record group.

    These entries should be strings that are previously contained in
    the 'FIELDS' class constant.

    While strictly no constraints on to which fields can be added here
    you should only choose those fields whose features could fit into
    a plaintext csv or similar format.

    """


    # valid debug modes
    DEBUG_MODES = (True, False,)

    def __init__(self,
                 min_num_walkers=Ellipsis,
                 max_num_walkers=Ellipsis,
                 debug_mode=False,
                 **kwargs):
        """Constructor for Resampler class

        Parameters
        ----------

        min_num_walkers : int or None or Ellipsis
            The minimum number of walkers allowed to have. None is
            unbounded, and Ellipsis preserves whatever number of
            walkers were given as input as the minimum.

        max_num_walkers : int or None or Ellipsis
            The maximum number of walkers allowed to have. None is
            unbounded, and Ellipsis preserves whatever number of
            walkers were given as input as the maximum.

        debug_mode : bool
            Expert mode stuff don't use unless you know what you are doing.


        """

        # the min and max number of walkers that can be generated in
        # resampling.

        # Ellipsis means to keep bound it by the number of
        # walkers given to the resample method (e.g. if
        # max_num_walkers == Ellipsis and min_num_walkers == 5 and
        # resample is given 10 then the max will be set to 10 for that
        # resampling and the min will always be 5. If both are
        # Ellipsis then the number of walkers is kept the same)

        # None means that there is no bound, e.g. max_num_walkers ==
        # None then there is no maximum number of walkers, however a
        # min_num_walkers of None in practice is 1 since there must
        # always be at least 1 walker

        if min_num_walkers not in (Ellipsis, None):
            if min_num_walkers < 1:
                raise ResamplerError("The minimum number of walkers should be at least 1")

        self._min_num_walkers = min_num_walkers
        self._max_num_walkers = max_num_walkers

        # this will be used to save the number of walkers given during
        # resampling, we initialize to None
        self._resampling_num_walkers = None

        # initialize debug mode
        self._debug_mode = False

        # set them to the args given
        self.set_debug_mode(debug_mode)


    @property
    def decision(self):
        """The decision class for this resampler."""
        return self.DECISION

    def resampling_field_names(self):
        """Access the class level FIELDS constant for this record group."""
        return self.RESAMPLING_FIELDS

    def resampling_field_shapes(self):
        """Access the class level SHAPES constant for this record group."""
        return self.RESAMPLING_SHAPES

    def resampling_field_dtypes(self):
        """Access the class level DTYPES constant for this record group."""
        return self.RESAMPLING_DTYPES

    def resampling_fields(self):
        """Returns a list of zipped field specs.

        Returns
        -------

        record_specs : list of tuple
            A list of the specs for each field, a spec is a tuple of
            type (field_name, shape_spec, dtype_spec)
        """
        return list(zip(self.resampling_field_names(),
                   self.resampling_field_shapes(),
                   self.resampling_field_dtypes()))

    def resampling_record_field_names(self):
        """Access the class level RECORD_FIELDS constant for this record group."""
        return self.RESAMPLING_RECORD_FIELDS

    def resampler_field_names(self):
        """Access the class level FIELDS constant for this record group."""
        return self.RESAMPLER_FIELDS

    def resampler_field_shapes(self):
        """Access the class level SHAPES constant for this record group."""
        return self.RESAMPLER_SHAPES

    def resampler_field_dtypes(self):
        """Access the class level DTYPES constant for this record group."""
        return self.RESAMPLER_DTYPES

    def resampler_fields(self):
        """Returns a list of zipped field specs.

        Returns
        -------

        record_specs : list of tuple
            A list of the specs for each field, a spec is a tuple of
            type (field_name, shape_spec, dtype_spec)
        """
        return list(zip(self.resampler_field_names(),
                   self.resampler_field_shapes(),
                   self.resampler_field_dtypes()))

    def resampler_record_field_names(self):
        """Access the class level RECORD_FIELDS constant for this record group."""
        return self.RESAMPLER_RECORD_FIELDS

    @property
    def is_debug_on(self):
        """ """
        return self._debug_mode

    def set_debug_mode(self, mode):
        """

        Parameters
        ----------
        mode

        Returns
        -------

        """

        if mode not in self.DEBUG_MODES:
            raise ValueError("debug mode, {}, not valid".format(mode))

        self._debug_mode = mode

        # if you want to use debug mode you have to have ipdb installed
        if self.is_debug_on:
            try:
                import ipdb
            except ModuleNotFoundError:
                raise ModuleNotFoundError("You must have ipdb installed to use the debug feature")

    def debug_on(self):
        """ """
        if self.is_debug_on:
            warn("Debug mode is already on")

        self.set_debug_mode(True)

    def debug_off(self):
        """ """
        if not self.is_debug_on:
            warn("Debug mode is already off")

        self.set_debug_mode(False)

    @property
    def max_num_walkers_setting(self):
        """The specification for the maximum number of walkers for the resampler."""
        return self._max_num_walkers

    @property
    def min_num_walkers_setting(self):
        """The specification for the minimum number of walkers for the resampler. """
        return self._min_num_walkers

    def max_num_walkers(self):
        """" Get the max number of walkers allowed currently"""

        # first check to make sure that a resampling is occuring and
        # we have a number of walkers to even reference
        if self._resampling_num_walkers is None:
            raise ResamplerError(
            "A resampling is currently not taking place so the"\
            " current number of walkers is not known.")

        # we are in a resampling so there is a current value for the
        # max number of walkers
        else:

            # if the max is None then there is no max number of
            # walkers so we just return None
            if self.max_num_walkers_setting is None:
                return None

            # if the max is Ellipsis then we just return what the
            # current number of walkers is
            elif self.max_num_walkers_setting is Ellipsis:
                return self._resampling_num_walkers

            # if it is not those then it is a hard number and we just
            # return it
            else:
                return self.max_num_walkers_setting

    def min_num_walkers(self):
        """" Get the min number of walkers allowed currently"""

        # first check to make sure that a resampling is occuring and
        # we have a number of walkers to even reference
        if self._resampling_num_walkers is None:
            raise ResamplerError(
            "A resampling is currently not taking place so the"\
            " current number of walkers is not known.")

        # we are in a resampling so there is a current value for the
        # min number of walkers
        else:

            # if the min is None then there is no min number of
            # walkers so we just return None
            if self.min_num_walkers_setting is None:
                return None

            # if the min is Ellipsis then we just return what the
            # current number of walkers is
            elif self.min_num_walkers_setting is Ellipsis:
                return self._resampling_num_walkers

            # if it is not those then it is a hard number and we just
            # return it
            else:
                return self.min_num_walkers_setting


    def _set_resampling_num_walkers(self, num_walkers):
        """Sets the concrete number of walkers constraints given a number of
        walkers and the settings for max and min.

        Parameters
        ----------
        num_walkers : int

        """

        # there must be at least 1 walker in order to do resampling
        if num_walkers < 1:
            raise ResamplerError("No walkers were given to resample")

        # if the min number of walkers is not dynamic check to see if
        # this number violates the hard boundary
        if self._min_num_walkers in (None, Ellipsis):
            self._resampling_num_walkers = num_walkers
        elif num_walkers < self._min_num_walkers:
            raise ResamplerError(
                "The number of walkers given to resample is less than the minimum")

        # if the max number of walkers is not dynamic check to see if
        # this number violates the hard boundary
        if self._max_num_walkers in (None, Ellipsis):
            self._resampling_num_walkers = num_walkers
        elif num_walkers < self._max_num_walkers:
            raise ResamplerError(
                "The number of walkers given to resample is less than the maximum")

    def _unset_resampling_num_walkers(self):

        self._resampling_num_walkers = None



    def _resample_init(self, walkers, **kwargs):
        """Common initialization stuff for resamplers.

        Sets the number of walkers in this round of resampling.

        Parameters
        ----------
        walkers : list of Walker objects

        """

        # first set how many walkers there are in this resampling
        self._set_resampling_num_walkers(len(walkers))

    def _resample_cleanup(self, **kwargs):
        """Common cleanup stuff for resamplers.

        Unsets the number of walkers for this round of resampling.

        """

        # unset the number of walkers for this resampling
        self._unset_resampling_num_walkers()


    @log_call(include_args=[],
              include_result=False)
    def resample(self, walkers, debug_mode=False):
        """Perform resampling on the set of walkers.

        Parameters
        ----------
        walkers : list of Walker objects
            The walkers that are to be resampled.

        debug_mode : bool
            Expert mode debugging setting, only forif you know exactly
            what you are doing.

        Returns
        -------

        resampled_walkers : list of Walker objects
            The set of resampled walkers

        resampling_data : list of dict of str: value
            A list of destructured resampling records from this round
            of resampling.

        resampler_data : list of dict of str: value
            A list of records recording how the state of the resampler
            was updated.

        """

        raise NotImplemented

        self._resample_init(walkers, debug_mode=debug_mode)


class NoResampler(Resampler):
    """The resampler which does nothing."""

    DECISION = NoDecision

    # must reset these when you change the decision
    RESAMPLING_FIELDS = DECISION.FIELDS + Resampler.CYCLE_FIELDS
    RESAMPLING_SHAPES = DECISION.SHAPES + Resampler.CYCLE_SHAPES
    RESAMPLING_DTYPES = DECISION.DTYPES + Resampler.CYCLE_DTYPES

    RESAMPLING_RECORD_FIELDS = DECISION.RECORD_FIELDS + Resampler.CYCLE_RECORD_FIELDS

    @log_call(include_args=[],
              include_result=False)
    def resample(self, walkers, **kwargs):

        self._resample_init(walkers=walkers)

        n_walkers = len(walkers)

        # the walker actions are all nothings with the same walker
        # index which is the default initialization
        resampling_data = self._init_walker_actions(n_walkers)

        # normally decide is only for a single step and so does not
        # include the step_idx, so we add this to the records, and
        # convert the target idxs and decision_id to feature vector
        # arrays
        for walker_idx, walker_record in enumerate(resampling_data):
            walker_record['walker_idx'] = np.array([walker_idx])
            walker_record['step_idx'] = np.array([0])
            walker_record['walker_idx'] = np.array([walker_record['walker_idx']])
            walker_record['decision_id'] = np.array([walker_record['decision_id']])
            walker_record['target_idxs'] = np.array([walker_record['walker_idx']])

        # we only have one step so our resampling_records are just the
        # single list of walker actions
        resampling_data = resampling_data

        # there is no change in state in the resampler so there are no
        # resampler records
        resampler_data = [{}]

        # the resampled walkers are just the walkers

        self._resample_cleanup(resampling_data=resampling_data,
                               resampler_data=resampler_data,
                               walkers=walkers)

        return walkers, resampling_data, resampler_data

    def _init_walker_actions(self, n_walkers):
        """Returns a list of default resampling records for a single
        resampling step.

        Parameters
        ----------

        n_walkers : int
            The number of walkers to generate records for

        Returns
        -------

        decision_records : list of dict of str: value
            A list of default decision records for one step of
            resampling.

        """
        # determine resampling actions
        walker_actions = [self.decision.record(
                              enum_value=self.decision.default_decision().value)
                    for i in range(n_walkers)]

        return walker_actions



"""This module sets up a random walk system then runs a simulation using
the given resampler. This profiles the performance of the resampler
using the data produced in the simulation.

The simulation data is saved using the WepyHDF5 reporter. To create a
WepyHDF5 reporter we create a dummy topology for the random walk
system. This dummy system consists of one atom with a position vector
in an N-dimensional space.

The resmapler quality is measured using the following values:

    - P(x): The average predicted probability at position x.

    - Accuracy: This is calculated using the target probability and predicted
    probabilities of walkers at position x.

   - Range: The maximum range that is observed by the given resampler is calculated
     by determining the largest x values visited along each dimension,
     then averaging them.

You can find detailed information about random walk parameters in the papers:

"WExplore: Hierarchical Exploration of High-Dimensional Spaces
Using the Weighted Ensemble Algorithm" and
"REVO: Resampling of Ensembles by Variation Optimization".

"""

import os
import sys
import json

import numpy as np
import pandas as pd
import h5py

import mdtraj as mdj

from wepy.resampling.resamplers.resampler import NoResampler
from wepy.work_mapper.mapper import Mapper
from wepy.reporter.hdf5 import WepyHDF5Reporter
from wepy.sim_manager import Manager
from wepy.walker import Walker, WalkerState
from wepy.runners.randomwalk import RandomWalkRunner, UNIT_NAMES
from wepy.hdf5 import WepyHDF5
from wepy.util.mdtraj import mdtraj_to_json_topology


PRECISION = 3

SAVE_FIELDS = ('positions',)
UNITS=UNIT_NAMES


np.set_printoptions(precision=PRECISION)



class RandomwalkProfiler(object):
    """A class to implement RandomWalkProfilier."""


    RANDOM_WALK_TEMPLATE=\
"""* Random walk simulation:
-Number of runs: {n_runs}
-Number of cycles: {n_cycles}
-Number of walkers:{n_walkers}
-Move-forward probability:{prob}
-Dimension:{dimension}
"""
    RUN_RESULTS_TEMPLATE=\
"""* Run {run_idx} results:
-Maximum range: {max_range}
-Maximum range of dimensions:
{max_dim_range}
-Accuracy: {accuracy}
-Average Predicted probability:
{predicted_probabilty}

"""

    def __init__(self, resampler=None, dimension=None, probility=0.25,
                 hdf5_filename='rw_results.wepy.h5',
                 reporter_filename='randomwalk.org'):
        """Constructor for RandomwalkProfiler.

        Parameters
        ----------
        resampler:
            The dimension of the random walk space.
             (Default = 2)

        probabilty: float
            "Probability" is defined here as the forward-move
             probability only. The backward-move probability is
             1-probability.
             (Default = 0.25)

        """

        assert resampler is not None,  "Resampler object must be given."
        self.resampler = resampler

        assert dimension is not None,  "The dimension of random walk  must be given."
        self.dimension = dimension

        self.probability = probility

        self.hdf5_filename = hdf5_filename

        self.reporter_filename = reporter_filename

    def generate_topology(self):
        """Creates a one-atom, dummy trajectory and topology for
        the randomwalk system using the mdtraj package.  Then creates a
        JSON format for the topology. This JSON string is used in making
        the WepyHDF5 reporter.

        Returns
        -------
        topology: str
            JSON string representing the topology of system being simulated.

        """
        n_atoms = 1
        data = []
        for i in range(n_atoms):
            data.append(dict(serial=i, name="H", element="H",
                             resSeq=i + 1, resName="UNK", chainID=0))

        data = pd.DataFrame(data)

        xyz = np.zeros((1, 1, 3))
        unitcell_lengths = 0.943 * np.ones((1, 3))
        unitcell_angles = 90 * np.ones((1, 3))

        top = mdj.Topology.from_dataframe(data, bonds=np.zeros((0, 2), dtype='int'))

        json_top_str = mdtraj_to_json_topology(top)

        return json_top_str


    def run(self, num_runs=1, num_cycles=200, num_walkers=100):
        """Runs a random walk simulation and profiles the resampler
        performance.

        Parameters
        ----------
        num_runs: int
            The number independet simulations.

        num_cycles: int
            The number of cycles that will be run in the simulation.

        num_walkers: int
            The number of walkers.

        """
        # set the random walk simulation repreter string
        randomwalk_string = self.RANDOM_WALK_TEMPLATE.format(
            n_runs=num_runs,
            n_cycles=num_cycles,
            n_walkers=num_walkers,
            prob=self.probability,
            dimension=self.dimension
            )

        # calls the runner
        self._run(num_runs, num_cycles, num_walkers)

        # calls the profiler
        self.analyse(randomwalk_string)


    def _run(self, num_runs, num_cycles, num_walkers):
        """Runs a random walk simulation.

        Parameters
        ----------
        num_runs: int
            The number independet simulations.

        num_cycles: int

            The number of cycles that will be run in the simulation.

        num_walkers: int
            The number of walkers.

        """

        print("Random walk simulation with: ")
        print("Dimension = {}".format(self.dimension))
        print("Probability = {}".format(self.probability))
        print("Number of Walkers = {}".format(num_walkers))
        print("Number of Cycles ={}".format(num_cycles))

        # set up initial state for walkers
        positions = np.zeros((1, self.dimension))

        init_state = WalkerState(positions=positions, time=0.0)


        # create list of init_walkers
        initial_weight = 1/num_walkers
        init_walkers = []

        init_walkers = [Walker(init_state, initial_weight)
                        for i in range(num_walkers)]

        # set up raunner for system
        runner = RandomWalkRunner(probability=self.probability)

        units = dict(UNIT_NAMES)
        # instantiate a revo unbindingboudaryconditiobs
        segment_length = 10

        # set up the reporter
        randomwalk_system_top_json = self.generate_topology()

        hdf5_reporter = WepyHDF5Reporter(file_path=self.hdf5_filename,
                                         mode='w',
                                         save_fields=SAVE_FIELDS,
                                         topology=randomwalk_system_top_json,
                                         resampler=self.resampler,
                                         units=dict(UNITS),
                                         n_dims=self.dimension)
        # running the simulation
        sim_manager = Manager(init_walkers,
                              runner=runner,
                              resampler=self.resampler,
                              work_mapper=Mapper(),
                              reporters=[hdf5_reporter])


        # run a simulation with the manager for n_steps cycles of length 1000 each
        steps = [segment_length for i in range(num_cycles)]
        ### RUN the simulation
        for run_idx in range(num_runs):
            print("Starting run: {}".format(run_idx))
            sim_manager.run_simulation(num_cycles, steps)
            print("Finished run: {}".format(run_idx))


        print("Finished Simulation")



    def kronecker_delta(self, x):
        """Implements the the Kronecker delta function.

        Parameters
        ----------
        x: int
            Input value of the function. Here, this is the random walk position.


        Returns
        -------
        y: int
            The output of the the Kronecker delta function.

        """

        if x == 0:
            return 1
        else:
            return 0

    # Measure accuracy
    def accuracy(self, x, Px):
        """Calculate the accuracy at position x.

        Parameters
        ----------
        x: int
            The position.

        Returns
        -------

        accuracy: float
            The value that specifies how accurate the resampler is at point x.
            The highest accuracy is achived when P(X) = Pt(x).


        """

        if Px == 0:
            return 0
        elif np.log(Px) > 2* np.log(self.Pt(x)):
            return 1 + np.abs( np.log(self.Pt(x)) - np.log(Px)) / np.log(self.Pt(x))
        else:
            return 0


    def Pt(self, x):
        """Calculate the target probability at position x.

        Parameters
        ----------

        x: int
            The position.

        Returns
        -------

        accuracy : float
            The value of the target probability when the
            forward-move probability p.

        """

        # p/q where q is (1-p)
        ratio = self.probability/(1-self.probability)

        if self.probability > (1-self.probability):
            return 1

        else:

            return (1 - ratio) * np.power(ratio, x)


    def get_max_range(self, wepy_h5, run_idx=0):
        """Finds the furthest range (position) visited by the walkers in the
        simulation.

        Parameters
        ----------
        wepy_h5: Hdf5 file object
            The hdf5 file that the simulation data is stored in.

        run_idx: int
            The index of the run.

        Returns
        -------

        max_range: int
            The maximum range that is visited by all walkers in
            all dimensions.

        max_dim_range: arraylike of shape (1, dimension)
            The miximum value of range in each dimension.

        """

        max_ranges = []
        num_cycles = wepy_h5.num_run_cycles(run_idx)
        num_walkers = wepy_h5.num_run_trajs(run_idx)

        for walker_idx in range(num_cycles):

            # makes the trace for the current walker
            walker_trace = [(walker_idx, cycle_idx) for cycle_idx in range(num_cycles)]

            # gets the data of given fields for the walker
            traj_data = wepy_h5.get_run_trace_fields(run_idx, walker_trace,
                                                     ['positions'])
            # gets the walker posiotions
            positions = traj_data['positions']

            max_ranges.append(np.max(positions, axis=0)[0])


        return np.max(np.array(max_ranges)), np.max(np.array(max_ranges), axis=0)



    def get_predicted_probability(self, wepy_h5, run_idx, max_range):
        """This projects all dimensions in the simulation on a
           1-dimensional space. It then calculates the predicted
           probability for all positions in in this 1D space in the
           range 0 to max_x.


        Parameters
        ----------
        walker: object implementing the Walker interface
            The individual walker for which dynamics will be propagated.

        segment_length : int
            The numerical value that specifies how much dynamical steps
            are to be run.

        Returns
        -------
        predicted_probabilty: list of floats
            The average predicted probability distribution.

        """

        predited_probibilty = np.zeros((max_range))

        num_cycles = wepy_h5.num_run_cycles(run_idx)
        num_walkers = wepy_h5.num_run_trajs(run_idx)

        for walker_idx in range(num_cycles):
            # makes the trace for the current walker
            walker_trace = [(walker_idx, cycle_idx) for cycle_idx in range(num_cycles)]
            # gets the data of given fields for the current walker
            traj_data = wepy_h5.get_run_trace_fields(run_idx, walker_trace, ['positions', 'weights'])

            weight = traj_data['weights']
            positions = traj_data['positions']

            # calculates for all x in range 0 to max_range
            for x in range(max_range):
                for cycle_idx in range(num_cycles):
                    for dim_idx in range(self.dimension):
                        predited_probibilty[x] += weight[walker_idx] * \
                        self.kronecker_delta(x - positions[walker_idx, 0, dim_idx])

        return 1/(num_cycles * self.dimension) * predited_probibilty

    def get_accuracy(self, predicted_probabilty):
        """Calculates the accuracy of the resampler for a random walk simulation
        run.

        Parameters
        ----------
        predicted_probabilty: list of floats
            Average predicted probability distribution.

        Returns
        -------
        accuracy: float
            The accuracy of the resampler determined using
            the accuracy equation.
        """


        accuracy_value = 0
        for x, px in enumerate(predicted_probabilty):
            accuracy_value += self.accuracy(x, px)

        return accuracy_value

    def analyse(self, randomwalk_string):
        """Calculates all quality metrics for the random walk simulation
        including pridicted probabilities, accuracy, and maximum
        average range.

        Parameters
        ----------
        walker: object implementing the Walker interface
            The individual walker for which dynamics will be propagated.

        Returns
        -------
        results: dict of str/arraylike

        """

        wepy_h5 = WepyHDF5(self.hdf5_filename, mode='r')
        wepy_h5.open()


        # get the number of euns
        num_runs = wepy_h5.num_runs


        results = []
        for run_idx in range(num_runs):

            max_range, max_dim_ranges = self.get_max_range(wepy_h5, run_idx)

            predicted_probabilty = self.get_predicted_probability(wepy_h5, run_idx,
                                           int(max_range))

            accuracy_value = self.get_accuracy(predicted_probabilty)

            run_results = {'max_range': max_range,
                       'max_dim_range': max_dim_ranges,
                       'predicted_probabilty': predicted_probabilty,
                       'accuracy': accuracy_value}

            results.append(run_results)

        self.write_report(randomwalk_string, results)

        return results


    def run_string(self, run_idx, run_results):
        """Creates a formated string for writing results of a randomwalk run.

        Parameters
        ----------
        run_idx: int
            The index of the run.

        Returns
        -------
        results: str
            The formated results.

        """

        run_result_string = self.RUN_RESULTS_TEMPLATE.format(
            run_idx=run_idx,
            max_range=run_results['max_range'],
            max_dim_range=np.array2string(run_results['max_dim_range'],
                                        separator=','),
            accuracy=round(run_results['accuracy'], PRECISION),

            predicted_probabilty=np.array2string(run_results['predicted_probabilty'],
                                                 separator=',')
            )


        return run_result_string



    def write_report(self, randomwalk_string, results):
        """Write the dashboard to the file."""

        mode = 'w'

        report_string = randomwalk_string

        for run_idx, run_results in enumerate(results):
            report_string += self.run_string(run_idx, run_results)

        with open(self.reporter_filename, mode=mode) as reporter_file:
            reporter_file.write(report_string)



"""Resampler framework and library.

This sub-package provides an interface for implementing new resamplers
that are able to reuse existing distance metrics and decision classes.

Resamplers minimally must implement a single method 'resample'.

"""



from pathlib import Path

import numpy as np

from wepy.hdf5 import WepyHDF5
from wepy.resampling.decisions.clone_merge import MultiCloneMergeDecision
from wepy.boundary_conditions.unbinding import UnbindingBC
from wepy.analysis.transitions import run_transition_probability_matrix
from wepy.analysis.network import MacroStateNetwork
from wepy.analysis.contig_tree import ContigTree

output_dir = Path('_output')
sim_dir = output_dir / 'we'

# Load wepy hdf5 file into python script
wepy_h5 = WepyHDF5(sim_dir / 'results.wepy.h5', mode = 'r+')
run_idx = 0
assg_key = 'rand_assg_idx'
n_classifications = 4
random_seed = 1

np.random.seed(random_seed)

# make random assignments

# observable function
def rand_assg(fields_d, *args, **kwargs):
    assignments = np.random.randint(0, n_classifications,
                                            size=fields_d['weights'].shape)
    return assignments

with wepy_h5:

    # compute this random assignment "observable"
    wepy_h5.compute_observable(
        rand_assg,
        ['weights'],
        (),
        save_to_hdf5=assg_key,
        return_results=False,
    )

# make a contig tree
contig_tree = ContigTree(
    wepy_h5,
    decision_class=MultiCloneMergeDecision,
    boundary_condition_class=UnbindingBC,
)


# make a state network using the key of the assignment and the
# file. Without edges this is just a collection of "macrostates"
# or groupings of microstates from the simulation data. We also
# set some sort of lag time to compute transition probabilities
# with

random_macrostates = MacroStateNetwork(contig_tree,
                                       assg_field_key="observables/{}".format(assg_key),
                                       transition_lag_time=3)

node_id = random_macrostates.node_ids[0]

# we can do things like make a trajectory in mdtraj and output as a
# dcd for a state
traj = random_macrostates.state_to_mdtraj(node_id)
traj.save_dcd(str(output_dir / "state.dcd".format(node_id)))

# we also can automatically compute the weights of the macrostates.
random_macrostates.set_macrostate_weights()

# this sets them as macrostate (node) attributes
print("node {} weight:".format(node_id))
print(random_macrostates.get_node_attribute(node_id, 'Weight'))

# we can also get a transition probability matrix from this
print(random_macrostates.probmat)


# furthermore you can write the network out to a GEXF file that
# can be used for visualization
random_macrostates.write_gexf(str(output_dir / "random_macrostates.csn.gexf"))



from copy import copy, deepcopy

class SimApparatus():
    """The simulation apparatus are the components needed for running a
    simulation without the initial conditions for starting the simulation.
    
    A runner is strictly necessary but a resampler and boundary
    conditions are not.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, filters):
        self._filters = deepcopy(filters)

    @property
    def filters(self):
        """ """
        return self._filters



class WepySimApparatus(SimApparatus):
    """ """

    RUNNER_IDX = 0
    BC_IDX = 1
    RESAMPLER_IDX = 2

    def __init__(self, runner, resampler=None, boundary_conditions=None):

        if resampler is None:
            raise ValueError("must provide a resampler")

        # add them in the order they are done in Wepy
        filters = [runner, boundary_conditions, resampler]

        super().__init__(filters)

    @property
    def runner(self):
        return self.filters[self.RUNNER_IDX]

    @property
    def boundary_conditions(self):
        return self.filters[self.BC_IDX]

    @property
    def resampler(self):
        return self.filters[self.RESAMPLER_IDX]


class SimSnapshot():
    """ """

    def __init__(self, walkers, apparatus):

        self._walkers = deepcopy(walkers)
        self._apparatus = deepcopy(apparatus)

    @property
    def walkers(self):
        """ """
        return self._walkers

    @property
    def apparatus(self):
        """ """
        return self._apparatus




"""Classes for workers and tasks for use with WorkerMapper. """

import multiprocessing as mp
import time
import logging
from wepy.work_mapper.mapper import ABCWorkerMapper, WrapperException, TaskException

# this whole thing should get refactored into a better name which
# should be something like ConsumerMapper because our workers act like
# consumers



# we can't move the WorkerMapper here until some of the pickles I have
# laying around don't expect it to be here. In the meantime, new
# software can expect it to be here so we import it here.
from wepy.work_mapper.mapper import WorkerMapper, Worker, WorkerException




"""Reference implementations, abstract base classes, and a production
ready worker style mapper for mapping runner dynamics to walkers for
wepy simulation cycles.

"""
import sys
import multiprocessing as mp
import queue as pyq
import traceback
import time
from warnings import warn
import signal

import logging
from eliot import start_action, log_call

PY_MAP = map

class ABCMapper(object):
    """Abstract base class for a Mapper."""

    def __init__(self, segment_func=None, **kwargs):
        """Constructor for the Mapper class. No arguments are required.

        Parameters
        ----------
        segment_func : callable, optional
            Set a default segment_func. Typically set at runtime.

        """

        self._func = segment_func

        self._attributes = kwargs

    @property
    def attributes(self):
        return self._attributes

    @attributes.getter
    def attributes(self, key):
        return self._attributes[key]

    @log_call(include_args=[],
              include_result=False)
    def init(self, segment_func=None, **kwargs):
        """Runtime initialization and setting of function to map over walkers.

        Parameters
        ----------
        segment_func : callable implementing the Runner.run_segment interface

        """


        if self.segment_func is not None and segment_func is not None:
            logging.info("overriding default segment_func {} with {}".format(self._func, segment_func))
            self._func = segment_func

        elif self.segment_func is None and segment_func is None:
            ValueError("segment_func must be given since no default specified")

        elif self.segment_func is None and segment_func is not None:
            self._func = segment_func


    @property
    def segment_func(self):
        """The function that will be called for new data in the `map` method."""
        return self._func

    @log_call(include_args=[],
              include_result=False)
    def cleanup(self, **kwargs):
        """Runtime post-simulation tasks.

        This is run either at the end of a successful simulation or
        upon an error in the main process of the simulation manager
        call to `run_cycle`.

        The Mapper class performs no actions here and all arguments
        are ignored.

        """

        # nothing to do
        pass

    @log_call(include_args=[],
              include_result=False)
    def map(self, *args, **kwargs):
        raise NotImplementedError



class Mapper(ABCMapper):
    """Basic non-parallel reference implementation of a mapper."""

    def __init__(self, segment_func=None, **kwargs):
        """Constructor for the Mapper class. No arguments are required.

        Parameters
        ----------
        segment_func : callable, optional
            Set a default segment_func. Typically set at runtime.

        """

        super().__init__(segment_func=segment_func, **kwargs)


        self._worker_segment_times = {0 : []}

    @log_call(include_args=[],
              include_result=False)
    def map(self, *args, **kwargs):
        """Map the 'segment_func' to args.

        Parameters
        ----------
        *args : list of list
            Each element is the argument to one call of 'segment_func'.

        Returns
        -------
        results : list
            The results of each call to 'segment_func' in the same order as input.

        Examples
        --------

        >>> Mapper(segment_func=sum).map([(0,1,2), (3,4,5)])
        [3, 12]

        """

        # expand the generators for the args and kwargs
        args = [list(arg) for arg in args]
        kwargs = {key : list(kwarg) for key, kwarg in kwargs.items()}

        segment_times = []
        results = []
        for arg_idx in range(len(args[0])):
            start = time.time()

            # get just the args for this call to func
            call_args = [arg[arg_idx] for arg in args]
            call_kwargs = {key : value[arg_idx] for key, value in kwargs.items()}

            # run the task, catch any errors and reraise as a
            # TaskException to satisfy the pattern
            try:

                result = self._func(*call_args, **call_kwargs)

            except Exception as task_exception:

                # get the traceback for the exception
                tb = sys.exc_info()[2]

                msg = "Exception '{}({})' caught in a task.".format(
                                       type(task_exception).__name__, task_exception)
                traceback_log_msg = \
                    """Traceback:
--------------------------------------------------------------------------------
{}
--------------------------------------------------------------------------------
                    """.format(''.join(traceback.format_exception(
                        type(task_exception), task_exception, tb)),
                    )

                logging.critical(msg + '\n' + traceback_log_msg)

                # raise a TaskException to distinguish it from the worker
                # errors with the metadata about the original exception

                raise TaskException("Error occured during task execution, recovery not possible.",
                                wrapped_exception=task_exception,
                                tb=tb)

            end = time.time()
            segment_time = end - start
            segment_times.append(segment_time)

            results.append(result)

        self._worker_segment_times[0] = segment_times

        return results

    @property
    def worker_segment_times(self):
        """The run timings for each segment for each walker.

        Returns
        -------
        worker_seg_times : dict of int : list of float
            Dictionary mapping worker indices to a list of times in
            seconds for each segment run.

        """
        return self._worker_segment_times


class Task(object):
    """Class that composes a function and arguments."""

    def __init__(self, func, *args, **kwargs):
        """Constructor for Task.

        Parameters
        ----------
        func : callable
            Function to be called on the arguments.

        *args
            The arguments to pass to func

        """
        self.args = args
        self.kwargs = kwargs
        self.func = func

    def __call__(self, **worker_kwargs):
        """Makes the Task itself callable."""

        # run the function passing in the args for running it and any
        # worker information in the worker kwargs.
        return self.func(*self.args, **self.kwargs, **worker_kwargs)

class WrapperException(Exception):
    """Exception used for wrapping another exception.

    Since tracebacks can't be pickled we format it and save that
    instead.

    """

    def __init__(self, message,
                 # must be kwargs so we can pickle it (I know weird...)
                 wrapped_exception=None,
                 tb=None):
        super().__init__(message)

        # save the exception with the traceback
        self.wrapped_exception = wrapped_exception
        self.formatted_tb = traceback.format_tb(tb)


class TaskException(WrapperException):
    pass

class ABCWorkerMapper(ABCMapper):

    def __init__(self,
                 num_workers=None,
                 segment_func=None,
                 proc_start_method='fork',
                 **kwargs):
        """Constructor for WorkerMapper.


        Parameters
        ----------
        num_workers : int
            The number of worker processes to spawn.

        segment_func : callable, optional
            Set a default segment_func. Typically set at runtime.

        proc_start_method : str or None
            A string indicating the type of process start method to
            use from python multiprocessing typically 'fork', 'spawn',
            or 'forkserver', or the platform default for None. See
            documentation. Generates a context with the method
            multiprocessing.get_context(proc_start_method) on `init`.

        """

        super().__init__(segment_func=segment_func, **kwargs)


        self._proc_start_method = proc_start_method

        self._num_workers = num_workers
        self._worker_segment_times = None

        if num_workers is not None:
            self._worker_segment_times = {i : [] for i in range(self.num_workers)}

    def init(self, num_workers=None, segment_func=None,
             **kwargs):
        """Runtime initialization and setting of function to map over walkers.

        Parameters
        ----------
        num_workers : int
            The number of worker processes to spawn

        segment_func : callable implementing the Runner.run_segment interface

        """

        super().init(segment_func=segment_func)

        # create the multiprocessing context to use for spawning
        # processes here
        self._mp_ctx = mp.get_context(method=self._proc_start_method)

        # the number of workers must be given here or set as an object attribute
        if num_workers is None and self.num_workers is None:
            raise ValueError("The number of workers must be given, received {}".format(num_workers))

        # if the number of walkers was given for this init() call use
        # that, otherwise we use the default that was specified when
        # the object was created
        elif num_workers is not None and self.num_workers is None:
            self._num_workers = num_workers

        # update the worker segment times
        self._worker_segment_times = {i : [] for i in range(self.num_workers)}


    def cleanup(self, **kwargs):

        # ALERT: is this all we need to do? I have a hunch there is
        # more caveats, but these context objects are not really
        # documented

        # make sure the context for this work mapper is destroyed
        del self._mp_ctx

    @property
    def num_workers(self):
        """The number of worker processes."""
        return self._num_workers

    @property
    def worker_segment_times(self):
        """The run timings for each segment for each walker.

        Returns
        -------
        worker_seg_times : dict of int : list of float
            Dictionary mapping worker indices to a list of times in
            seconds for each segment run.

        """
        return self._worker_segment_times

    def _make_task(self, *args, **kwargs):
        """Generate a task from 'segment_func' attribute.

        Similar to partial evaluation (or currying).

        Args will be eventually used as the arguments to the call of
        'segment_func' by the worker processes when they receive the
        task from the queue.

        Returns
        -------
        task : Task object

        """
        return Task(self._func, *args, **kwargs)


# ----------------------------------
# everything below this logically belongs in worker.py and should be imported from there


class WorkerException(WrapperException):
    pass

class WorkerKilledError(ChildProcessError):
    pass


# TODO: move this class to the wepy.work_mapper.worker class where it
# belongs. It shouldn't be in this namespace, but we will leave it
# here. Furthermore I would like to rename it since we now have
# different worker mapper implementations with different concurrency
# models
class WorkerMapper(ABCWorkerMapper):
    """Work mapper implementation using multiple worker processes and task
    queue.

    Uses the python multiprocessing module to spawn multiple worker
    processes which watch a task queue of walker segments.
    """

    def __init__(self,
                 num_workers=None,
                 worker_type=None,
                 worker_attributes=None,
                 segment_func=None,
                 **kwargs):
        """Constructor for WorkerMapper.


        Parameters
        ----------
        num_workers : int
            The number of worker processes to spawn.

        worker_type : callable, optional
            Callable that generates an object implementing the Worker
            interface, typically a type from a Worker class.

        worker_attributes : dictionary
            A dictionary of values that are passed to the worker
            constructor as key-word arguments.

        segment_func : callable, optional
            Set a default segment_func. Typically set at runtime.

        """

        super().__init__(num_workers=num_workers,
                         segment_func=segment_func,
                         **kwargs)

        # since the workers will be their own process classes we
        # handle this data

        # attributes that will be passed to the worker constructors
        if worker_attributes is not None:
            self._worker_attributes = worker_attributes
        else:
            self._worker_attributes = {}

        # choose the type of the worker
        if worker_type is None:
            self._worker_type = Worker
            warn("worker_type not given using the default base class")
            logging.warn("worker_type not given using the default base class")
        else:
            self._worker_type = worker_type

    @property
    def worker_type(self):
        """The callable that generates a worker object.

        Typically this is just the type from the class definition of
        the Worker where the constructor is called.

        """
        return self._worker_type


    def init(self, num_workers=None, segment_func=None,
             **kwargs):
        """Runtime initialization and setting of function to map over walkers.

        Parameters
        ----------
        num_workers : int
            The number of worker processes to spawn

        segment_func : callable implementing the Runner.run_segment interface

        """

        super().init(num_workers=num_workers,
                     segment_func=segment_func,
                     **kwargs)

        manager = self._mp_ctx.Manager()

        # Establish communication queues

        # A queue for errors
        self._exception_queue = manager.Queue()

        # queue for the tasks we know the batch size so we don't need
        # a JoinableQueue
        self._task_queue = manager.Queue()

        # results queue
        self._result_queue = manager.Queue()

        # use pipes for communication channels between this parent
        # process and the children for sending specific interrupts
        # such as the signal to kill them. Note that the clean way to
        # end the process is to send poison pills on the task queue,
        # this is for other stuff. IRQ is a common abbreviation for
        # interrupts
        self._irq_parent_conns = []

        # Start workers, giving them all the queues
        self._workers = []
        for i in range(self.num_workers):

            # make a pipe to communicate with this worker for the int
            parent_conn, child_conn = self._mp_ctx.Pipe()
            self._irq_parent_conns.append(parent_conn)

            # create the worker giving it all of the communication
            # channels
            worker = self.worker_type(i,
                                      self._task_queue,
                                      self._result_queue,
                                      self._exception_queue,
                                      child_conn,
                                      mapper_attributes=self._attributes,
                                      **self._worker_attributes)
            self._workers.append(worker)

        # start the worker processes
        for worker in self._workers:
            worker.start()

            logging.info("Worker process started as name: {}; PID: {}".format(worker.name,
                                                                              worker.pid))

        # now that we have started the processes register the handler
        # for SIGTERM signals that will clean up our children cleanly
        signal.signal(signal.SIGTERM, self._sigterm_shutdown)

    def _sigterm_shutdown(self, signum, frame):

        logging.critical("Received external SIGTERM, forcing shutdown.")

        self.force_shutdown()

    def force_shutdown(self, **kwargs):

        logging.critical("Forcing shutdown")

        # our primary job is to shut down all of the running processes
        # without just shutting down the queues and breaking the pipes

        # to do this we send the kill signals to them on the kill
        # channel.

        for worker_idx, worker in enumerate(self._workers):

            logging.critical("Sending SIGTERM message on {} to worker {}".format(
                self._irq_parent_conns[worker_idx].fileno(), worker_idx))

            # send a kill message to the worker
            self._irq_parent_conns[worker_idx].send(signal.SIGTERM)

        logging.critical("All kill messages sent to workers")



        # check that all have exited
        alive_workers = [worker.is_alive() for worker in self._workers]
        worker_acks = {}
        worker_exitcodes = {}
        premature_exit = False
        while any(alive_workers) and not premature_exit:

            for worker_idx, worker in enumerate(self._workers):

                # ignore already known dead workers
                if not alive_workers[worker_idx]:
                    continue

                if worker.is_alive():

                    # if it is still alive and we have an ack from it
                    # just terminate. There is a bug in the code and
                    # is out of our control
                    if worker_idx in worker_acks:
                        logging.debug("Ack received from {} but has not shut down".format(worker.name))
                        premature_exit = True

                    # otherwise we need to try and receive the ack
                    elif self._irq_parent_conns[worker_idx].poll(1):

                        # receive the acknowledgement
                        ack = self._irq_parent_conns[worker_idx].recv()

                        logging.debug("Received {} acknowledgement from {}".format(ack,
                                                                                   worker.name))

                        # make sure the ack is affirmative
                        if ack is True:
                            worker_acks[worker_idx] = ack

                        # if it is an exeption wrap it as a worker
                        # error and use the os to kill the process
                        elif issubclass(type(ack), Exception):

                            # wrap it as a worker exception
                            exception = WorkerException(wrapped_exception=ack)
                            worker_acks[worker_idx] = exception

                            logging.critical("{} not responding, terminating with SIGTERM".format(
                                worker.name))

                            worker.terminate()


                else:
                    alive_workers[worker_idx] = False
                    worker_exitcodes[worker_idx] = worker.exitcode

        if any(alive_workers):
            logging.critical("Terminating main process with running workers {}".format(
                ','.join([str(worker_idx) for worker_idx in range(len(self._workers))
                          if alive_workers[worker_idx]])))




    def cleanup(self, **kwargs):
        """Runtime post-simulation tasks.

        This is run either at the end of a successful simulation or
        upon an error in the main process of the simulation manager
        call to `run_cycle`.

        The Mapper class performs no actions here and all arguments
        are ignored.

        """

        super().cleanup(**kwargs)

        # send poison pills (Stop signals) to the queues to stop them in a nice way
        # and let them finish up
        for i in range(self.num_workers):
            self._task_queue.put((None, None))

        # delete the queues and workers
        self._task_queue = None
        self._result_queue = None
        self._workers = None


    def map(self, *args, **kwargs):
        # docstring in superclass

        map_process = self._mp_ctx.current_process()
        logging.info("Mapping from process {}; PID {}".format(map_process.name, map_process.pid))

        # make tuples for the arguments to each function call
        task_args = zip(*args)
        kwargs = {key : list(kwarg) for key, kwarg in kwargs.items()}

        num_tasks = len(args[0])
        # Enqueue the jobs
        for task_idx, task_arg in enumerate(task_args):

            task_kwargs = {key : value[task_idx] for key, value in kwargs.items()}

            # a task will be the actual task and its task idx so we can
            # sort them later
            self._task_queue.put((task_idx, self._make_task(*task_arg, **task_kwargs)))


        logging.info("Waiting for tasks to be run")

        # poll the exception and result queues for results
        n_results_left = num_tasks
        results = []
        while n_results_left > 0:

            # first check if any errors came back but don't wait,
            # since the methods for querying whether it is empty or
            # not are not reliable we just try and if we don't get
            # anything we will come back around
            try:
                proc_name, pid, exception = self._exception_queue.get_nowait()
            except pyq.Empty:
                pass

            else:

                logging.error("Exception occured in process {}; pid {}.".format(
                    proc_name, pid))

                # we can handle Task and Worker exceptions differently
                if type(exception) == TaskException:

                    logging.critical("Exception encountered in a task which is unrecoverable."
                                "You will need to reconfigure your components in a stable manner.")


                    self.force_shutdown()

                    logging.critical("Shutdown complete.")
                    raise exception

                elif type(exception) == WorkerException:

                    # we make just an error message to say that errors
                    # in the worker may be due to the network or
                    # something and could recover
                    logging.error("Exception encountered in the work mapper worker process."
                                  "Recovery possible, see further messages.")

                    # However, the current implementation doesn't
                    # support retries or whatever so we issue a
                    # critical log informing that it has been elevated
                    # to critical and will force shutdown
                    logging.critical("Worker error mode resiliency not supported at this time."
                                     "Performing force shutdown and simulation ending.")

                    self.force_shutdown()

                    logging.critical("Shutdown complete.")
                    raise exception

                else:
                    logging.critical("Unknown exception encountered.")

                    self.force_shutdown()

                    logging.critical("Shutdown complete.")

                    raise exception


            # attempt to get something off of the results queue
            try:
                result = self._result_queue.get_nowait()
            except pyq.Empty:
                pass

            # if we get something handle it
            else:

                logging.info("Retrieved result: {}".format(result))
                results.append(result)

                # reduce the counter so we know when we are done
                n_results_left -= 1

        # sort the results according to their task_idx
        results.sort()

        # save the task run times, so they can be accessed if desired,
        # after clearing the task times from the last mapping

        # DEBUG: removing this because it should be set on init()
        #self._worker_segment_times = {i : [] for i in range(self.num_workers)}

        for task_idx, worker_idx, task_time, result in results:
            self._worker_segment_times[worker_idx].append(task_time)

        # then just return the values of the function
        return [result for task_idx, worker_idx, task_time, result in results]


# same for the worker in terms of refactoring
class Worker(mp.Process):
    """Worker process.

    This is a subclass of process with an overriden `__init__`
    constructor that will automatically generate the Process.

    When this class is constructed a new process will be formed.

    """

    NAME_TEMPLATE = "Worker-{}"
    """A string formatting template to identify worker processes in
    logs. The field will be filled with the worker index."""

    def __init__(self, worker_idx,
                 task_queue,
                 result_queue,
                 exception_queue,
                 interrupt_connection,
                 mapper_attributes=None,
                 **kwargs):
        """Constructor for the Worker class.

        Parameters
        ----------
        worker_idx : int
            The index of the worker. Should be unique.

        task_queue : multiprocessing.JoinableQueue
            The shared task queue the worker will watch for new tasks to complete.

        result_queue : multiprocessing.Queue
            The shared queue that completed task results will be placed on.

        interrupt_connection : multiprocessing.Connection
            One end of a pipe to listen for messages specific to this worker.

        mapper_attributes : None or dict
            A dictionary of the attributes of the mapper for reference in workers.

        kwargs :
            The worker specific attributes

        """

        # call the Process constructor
        mp.Process.__init__(self, name=self.NAME_TEMPLATE.format(worker_idx))

        self._exception_queue = exception_queue
        self._exception = None
        self._traceback = None

        # the queue that will trigger a shutdown in the event of failure
        self._irq_channel = interrupt_connection

        # also register the SIGTERM signal handler for graceful
        # shutdown with reporting to mapper
        signal.signal(signal.SIGTERM, self._sigterm_shutdown)

        self._worker_idx = worker_idx

        self._mapper_attributes = mapper_attributes

        # set all the kwargs into an attributes dictionary
        self._attributes = kwargs

        # the queues for work to be done and work done
        self._task_queue = task_queue
        self._result_queue = result_queue

        logging.debug("{} process created".format(self.name))


    @property
    def worker_idx(self):
        """Dictionary of attributes of the worker."""
        return self._worker_idx

    @property
    def attributes(self):
        """Dictionary of attributes of the worker."""
        return self._attributes

    @property
    def mapper_attributes(self):
        """Dictionary of attributes of the worker."""
        return self._mapper_attributes

    def run(self):

        logging.debug("{}: starting to run".format(self.name))

        # try to run the worker and it's task, except either class of
        # error that can come from it either from the worker
        # (WorkerException) or the task (TaskException) and communicate it
        # back to the main process

        # if we get an exception there is some cleanup logic
        run_exception = None

        try:

            # run the worker, which will retrieve its task from the
            # queue attempt to run the task, and if it succeeds will
            # put the results on the result queue, if the task fails
            # it will catch it and wrap it as a task exception
            self._run_worker()

        except TaskException as task_exception:

            logging.error("{}: TaskException caught".format(self.name))

            run_exception = task_exception

        # anything else is considered a WorkerException so take the
        # original exception and generate a worker exception from that
        except Exception as exception:

            logging.debug("{}: WorkerError caught".format(self.name))


            # get the traceback
            tb = sys.exc_info()[2]

            msg = "Exception '{}({})' caught in a worker.".format(
                                   type(exception).__name__, exception)
            traceback_log_msg = \
                """Traceback:
--------------------------------------------------------------------------------
{}
--------------------------------------------------------------------------------
                """.format(''.join(traceback.format_exception(
                    type(exception), exception, tb)),
                )

            logging.error("{}:".format(self.name) + msg + '\n' + traceback_log_msg)

            # raise a TaskError to distinguish it from the worker
            # errors with the metadata about the original exception

            worker_exception = WorkerException(
                "Error occured during worker execution.",
                wrapped_exception=exception,
                tb=tb)

            run_exception = worker_exception

        # raise worker_exception
        if run_exception is not None:

            logging.debug("{}: Putting exception on exception queue".format(self.name))

            # then put the exception and the traceback onto the queue
            # so we can communicate back to the parent process
            try:
                self._exception_queue.put((self.name, self.pid, run_exception))
            except BrokenPipeError as exc:
                logging.error(
                    "Pipe is broken indicating the root process has already exited:\n{}".format(
                        exc))

            # TODO: not sure if this is good or not
            # then reraise the exception so it can be caught
            # raise run_exception

    def _sigterm_shutdown(self, signum, frame):

        logging.debug("Received external SIGTERM kill command.")

        logging.debug("Alerting mapper that this will be honored.")

        # send an error to the mapper that the worker has been killed
        self._irq_channel.send(WorkerKilledError(
            "{} (pid: {}) killed by external SIGTERM signal".format(self.name, self.pid)))

        logging.debug("Acknowledgment sent")

        logging.debug("Shutting down process")

    def _shutdown(self):

        logging.debug("Received SIGTERM kill command from mapper")

        logging.debug("Acknowledging kill request will be honored")

        # report back that we are shutting down with a True
        self._irq_channel.send(True)

        logging.debug("Acknowledgment sent")

        logging.debug("Shutting down process")


    def _run_worker(self):

        # run the logic associated with communication and liveness of
        # the worker process itself, this is not necessarily a fatal
        # (critical) error and restarting a worker might resolve the
        # problem. This calls the _run_task method though which is
        # always critical since the logic in the code cannot be
        # disputed

        # TODO remove when confirmed that this works
        # worker_process = mp.current_process()
        logging.info("{}: Worker process started as name: {}; PID: {}".format(
            self.name, self.name, self.pid))

        while True:

            # check to see if there is any signals in the interrupt channel
            if self._irq_channel.poll():
                # get the message
                message = self._irq_channel.recv()

                logging.debug("{}: Received message from mapper on filehandle {}: {}".format(
                    self.name, self._irq_channel.fileno(), message))

                # handle the message

                # a SIGTERM is a signal to kill the process
                # unconditionally
                if message is signal.SIGTERM:

                    self._shutdown()

                    # break from the event (while) loop and shut down
                    break

                # anything is not recognized and we will continue and
                # report back that we don't recognize the message with
                # a ValueError object
                else:
                    logging.error("{}: Message not recognized, continuing operations and"
                                  " sending error to mapper".format(self.name))
                    self._irq_channel.send(
                        ValueError(
                            "Message: {} not recognized continuing operations".format(
                                message)))

            # get the next task
            try:
                 task_idx, next_task = self._task_queue.get(block=False, timeout=None)

                 logging.debug("{}: Got task {}".format(self.name, task_idx))

            except pyq.Empty:

                task_idx = None
                next_task = Ellipsis


            # # check for the poison pill which is the signal to stop
            if next_task is None:

                logging.info('{}: received {} {}: FINISHED'.format(
                    self.name, task_idx, next_task))

                # TODO remove since we aren't using joinble queue anymore
                # mark the poison pill task as done
                #self.task_queue.task_done()

                # and exit the loop
                break

            # only execute this if a task was actually receieved from
            # the queue; an Ellipsis indicates continue the loop
            elif next_task is not Ellipsis:

                logging.info('{}; task_idx : {}; args : {} '.format(
                    self.name, task_idx, next_task.args))

                # run the task
                start = time.time()

                answer = self._run_task(next_task)

                end = time.time()
                task_time = end - start

                logging.info('{}: task_idx : {}; COMPLETED in {} s'.format(
                    self.name, task_idx, task_time))

                # put the results into the results queue with it's task
                # index so we can sort them later
                self._result_queue.put((task_idx, self.worker_idx, task_time, answer))

    def run_task(self, task):
        """Actually executes the task.

        This default runner simply executes the task thunk.

        This can be customized by subclasses in order to allow for
        injection of worker specific data.

        Parameters
        ----------
        task : Task object
            The partially evaluated task; function plus arguments

        Returns
        -------
        task_result
            Results of running the task.

        """

        return task()


    def _run_task(self, task):
        """Runs the given task and returns the results.

        This manages handling exceptions and tracebacks from the
        actual `run_task` function which is intended to be specialized
        by different workers to inject worker specific arguments to
        tasks. Such as node and device identification.

        Parameters
        ----------
        task : Task object
            The partially evaluated task; function plus arguments

        Returns
        -------
        task_result
            Results of running the task.

        """

        logging.info("Running task")
        try:
            return self.run_task(task)

        except Exception as task_exception:

            # get the traceback for the exception
            tb = sys.exc_info()[2]

            msg = "Exception '{}({})' caught in a task.".format(
                                   type(task_exception).__name__, task_exception)
            traceback_log_msg = \
                """Traceback:
--------------------------------------------------------------------------------
{}
--------------------------------------------------------------------------------
                """.format(''.join(traceback.format_exception(
                    type(task_exception), task_exception, tb)),
                )

            logging.critical(msg + '\n' + traceback_log_msg)

            # raise a TaskException to distinguish it from the worker
            # errors with the metadata about the original exception

            raise TaskException("Error occured during task execution, recovery not possible.",
                            wrapped_exception=task_exception,
                            tb=tb)



from pint import UnitRegistry
from jinja2 import Template

from wepy.reporter.dashboard import BCDashboardSection

# initialize the unit registry
units = UnitRegistry()


class ReceptorBCDashboardSection(BCDashboardSection):


    BC_SECTION_TEMPLATE = \
"""

Boundary Condition: {{ name }}

Total Number of Dynamics segments: {{ total_n_walker_segments }}

Total Number of Warps: {{ total_crossings }}

Cumulative Boundary Crossed Weight: {{ total_unbound_weight }}

Rate (1/τ): {{ rate }}

Parameters:

{{ parameters }}


** Warping Log

{{ warping_log }}

"""


    def __init__(self, **kwargs):

        super().__init__(**kwargs)


    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        # since there is only one boundary to cross here we don't
        # really have to do any special reporting for meaningful
        # boundaries. So we just use the standard one.


        # we calculate the non-dimensional rate in terms of the cycle
        # numbers, then you would just have to multiply that number by
        # 1/time_per_cycle. In order to handle variable number of
        # walkers we just keep track of all of the segments that were
        # run total by keeping a running tally of how many
        # 'new_walkers' are received in the 'total_n_walkers'
        # attribute

        rate = self.total_crossed_weight / self.total_n_walker_segments

        new_fields = {
            'parameters' : '',
            'rate' : rate,
        }


        # combine the superclass fields with the fields here,
        # overwriting them from the superclass if they were redefined
        # explicitly
        fields.update(new_fields)

        return fields



class UnbindingBCDashboardSection(ReceptorBCDashboardSection):

    RECEPTOR_PARAMETERS = \
"""
Cutoff Distance: {{ cutoff_distance }}
"""

    def __init__(self, bc=None,
                 cutoff_distance=None,
                 **kwargs):

        if 'name' not in kwargs:
            kwargs['name'] = 'UnbindingBC'

        super().__init__(bc=bc,
                         cutoff_distance=cutoff_distance,
                         **kwargs)


        if bc is not None:
            self.cutoff_distance = bc.cutoff_distance
        else:
            assert cutoff_distance is not None, \
                "If no bc is given must give parameters: cutoff_distance"
            self.cutoff_distance = cutoff_distance

    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        parameters_str = Template(self.RECEPTOR_PARAMETERS).render(
            cutoff_distance=self.cutoff_distance,
        )

        new_fields = {
            'parameters' : parameters_str,
        }

        return fields

class RebindingBCDashboardSection(ReceptorBCDashboardSection):

    RECEPTOR_PARAMETERS = \
"""
Cutoff RMSD: {{ cutoff_rmsd }}
"""

    def __init__(self, bc=None,
                 cutoff_rmsd=None,
                 **kwargs):

        if 'name' not in kwargs:
            kwargs['name'] = 'RebindingBC'

        super().__init__(bc=bc,
                         cutoff_rmsd=cutoff_rmsd,
                         **kwargs)

        if bc is not None:
            self.cutoff_rmsd = bc.cutoff_rmsd
        else:
            assert cutoff_rmsd is not None, \
                "If no bc is given must give parameters: cutoff_rmsd"
            self.cutoff_rmsd = cutoff_rmsd

    def gen_fields(self, **kwargs):

        fields = super().gen_fields(**kwargs)

        parameters_str = Template(self.RECEPTOR_PARAMETERS).render(
            cutoff_rmsd=self.cutoff_rmsd,
        )

        new_fields = {
            'parameters' : parameters_str,
        }

        return fields



